
from os import getenv, environ
from os.path import exists, join, expanduser
from random import seed, shuffle, sample, randint, uniform
from itertools import islice as head
from subprocess import run

from tqdm.notebook import tqdm as log_progress

import torch
from torch import optim

from nerus import load_nerus

from navec import Navec

from slovnet.s3 import S3
from slovnet.io import (
    format_jl,
    parse_jl,

    load_gz_lines,
    dump_gz_lines
)
from slovnet.board import (
    TensorBoard,
    LogBoard,
    MultiBoard
)
from slovnet.const import (
    TRAIN, TEST,
    CUDA0, PAD,
    WORD, SHAPE, TAG
)
from slovnet.token import tokenize
from slovnet.shape import SHAPES

from slovnet.markup import (
    MorphToken,
    MorphMarkup,
    show_morph_markup
)
from slovnet.model.emb import (
    Embedding,
    NavecEmbedding
)
from slovnet.model.tag import (
    TagEmbedding,
    TagEncoder,
    MorphHead,
    Morph
)
from slovnet.loss import flatten_cross_entropy
from slovnet.vocab import Vocab
from slovnet.encoders.tag import TagTrainEncoder
from slovnet.score import (
    MorphBatchScore,
    MorphScoreMeter,
    score_morph_batch
)

from slovnet.exec.pack import (
    Meta,
    DumpPack
)
from slovnet import api


DATA_DIR = 'data'
MODEL_DIR = 'model'
NAVEC_DIR = 'navec'
RAW_DIR = join(DATA_DIR, 'raw')
S3_DIR = '06_morph'

RAW_NERUS = join(RAW_DIR, 'nerus_lenta.conllu.gz')
NERUS_TOTAL = 739346

NERUS = join(DATA_DIR, 'nerus.jl.gz')
S3_NERUS = join(S3_DIR, NERUS)

NAVEC_URL = 'https://storage.yandexcloud.net/natasha-navec/packs/navec_news_v1_1B_250K_300d_100q.tar'
NAVEC = join(NAVEC_DIR, 'navec_news_v1_1B_250K_300d_100q.tar')

TAGS_VOCAB = join(MODEL_DIR, 'tags_vocab.txt')
MODEL_SHAPE = join(MODEL_DIR, 'shape.pt')
MODEL_ENCODER = join(MODEL_DIR, 'encoder.pt')
MODEL_MORPH = join(MODEL_DIR, 'morph.pt')

S3_TAGS_VOCAB = join(S3_DIR, TAGS_VOCAB)
S3_MODEL_SHAPE = join(S3_DIR, MODEL_SHAPE)
S3_MODEL_ENCODER = join(S3_DIR, MODEL_ENCODER)
S3_MODEL_MORPH = join(S3_DIR, MODEL_MORPH)

ID = 'slovnet_morph_news_v1'
PACK = ID + '.tar'
S3_PACK = join('packs', PACK)

BOARD_NAME = getenv('board_name', '06_morph')
RUNS_DIR = 'runs'

TRAIN_BOARD = '01_train'
TEST_BOARD = '02_test'

SEED = int(getenv('seed', 65))
DEVICE = getenv('device', CUDA0)

SHAPE_DIM = int(getenv('shape_dim', 30))
LAYERS_NUM = int(getenv('layers_num', 3))
LAYER_DIM = int(getenv('layer_dim', 64))
KERNEL_SIZE = int(getenv('kernel_size', 3))

LR = float(getenv('lr', 0.0033))
LR_GAMMA = float(getenv('lr_gamma', 0.9))
EPOCHS = int(getenv('epochs', 5))

LAYER_DIMS = [
    LAYER_DIM * 2**_
    for _ in reversed(range(LAYERS_NUM))
]


def adapt_markup(record):
    return MorphMarkup([
        MorphToken(_.text, _.pos, _.feats)
        for _ in record.tokens
    ])


def process_batch(model, criterion, batch):
    input, target = batch

    pred = model(input.word_id, input.shape_id)
    loss = criterion(pred, target)

    pred = model.morph.decode(pred)

    return batch.processed(loss, pred)




import pytest

from slovnet.shape import (
    X, x, xx, XX, Xx, Xx_Xx, OTHER,
    RU, EN, NUM, PUNCT,

    word_shape,
    format_shape as s
)
from slovnet.token import tokenize


TESTS = [
    [
        'В',
        [s(RU, X)],
    ],
    [
        'ИЛ-2',
        [s(RU, XX)],
    ],
    [
        '105г.',
        [NUM, s(RU, x), s(PUNCT, '.')]
    ],
    [
        'Pal-Yz',
        [s(EN, Xx_Xx)]
    ],
    [
        'и Я-ДаА',
        [s(RU, x), s(RU, OTHER)]
    ],
    [
        'Прибыл на I@',
        [s(RU, Xx), s(RU, xx), s(EN, X), s(PUNCT, '@')]
    ],
    [
        'и -‐',
        [s(RU, x), s(PUNCT, OTHER)]
    ]
]


@pytest.mark.parametrize('test', TESTS)
def test_shape(test):
    text, etalon = test
    tokens = tokenize(text)
    guess = [word_shape(_.text) for _ in tokens]
    assert guess == etalon




from razdel import tokenize as tokenize_

from .record import Record


class Token(Record):
    __attributes__ = ['start', 'stop', 'text']


def tokenize(text):
    for token in tokenize_(text):
        yield Token(
            token.start,
            token.stop,
            token.text
        )


def find_tokens(text, chunks):
    offset = 0
    for chunk in chunks:
        start = text.find(chunk, offset)
        stop = start + len(chunk)
        yield Token(start, stop, chunk)
        offset = stop




try:
    from ipymarkup import (
        show_span_box_markup,
        show_dep_markup
    )
except ImportError:
    pass

from .record import Record
from .bio import (
    spans_bio,
    bio_spans
)
from .token import find_tokens
from .sent import sentenize
from .span import (
    Span,
    envelop_spans,
    offset_spans
)
from .conll import (
    format_conll_tag,
    parse_conll_tag
)


########
#
#   SPAN
#
#######


class SpanMarkup(Record):
    __attributes__ = ['text', 'spans']
    __annotations__ = {
        'spans': [Span]
    }

    @property
    def sents(self):
        for sent in sentenize(self.text):
            spans = envelop_spans(sent, self.spans)
            spans = offset_spans(spans, -sent.start)
            yield SpanMarkup(sent.text, list(spans))

    def to_bio(self, tokens):
        tags = spans_bio(tokens, self.spans)
        words = [_.text for _ in tokens]
        return BIOMarkup.from_tuples(zip(words, tags))


def show_span_markup(markup):
    show_span_box_markup(markup.text, markup.spans)


########
#
#   TAG
#
######


class TagToken(Record):
    __attributes__ = ['text', 'tag']


class TagMarkup(Record):
    __attributes__ = ['tokens']
    __annotations__ = {
        'tokens': [TagToken]
    }

    @property
    def words(self):
        return [_.text for _ in self.tokens]

    @property
    def tags(self):
        return [_.tag for _ in self.tokens]

    @classmethod
    def from_tuples(cls, tuples):
        return cls([
            TagToken(word, tag)
            for word, tag in tuples
        ])


class BIOMarkup(TagMarkup):
    def to_span(self, text):
        tokens = find_tokens(text, self.words)
        spans = list(bio_spans(tokens, self.tags))
        return SpanMarkup(text, spans)


########
#
#   MORPH
#
########


class MorphToken(TagToken):
    __attributes__ = ['text', 'pos', 'feats']

    @property
    def tag(self):
        return format_conll_tag(self.pos, self.feats)


class MorphMarkup(TagMarkup):
    __attributes__ = ['tokens']
    __annotations__ = {
        'tokens': [MorphToken]
    }

    @classmethod
    def from_tuples(cls, tuples):
        tokens = []
        for word, tag in tuples:
            pos, feats = parse_conll_tag(tag)
            tokens.append(MorphToken(word, pos, feats))
        return cls(tokens)


def format_morph_markup(markup, size=20):
    for word, tag in zip(markup.words, markup.tags):
        word = word.rjust(size)
        yield '%s %s' % (word, tag)


def show_morph_markup(markup):
    for line in format_morph_markup(markup):
        print(line)


def format_morph_markup_diff(a, b, size=20):
    for word, a_token, b_token in zip(a.words, a.tokens, b.tokens):
        word = word.rjust(size)
        a_tag = format_conll_tag(a_token.pos, a_token.feats)
        yield '%s   %s' % (word, a_tag)
        if a_token != b_token:
            word = ' ' * size
            b_tag = format_conll_tag(b_token.pos, b_token.feats)
            yield '%s ! %s' % (word, b_tag)


def show_morph_markup_diff(a, b):
    for line in format_morph_markup_diff(a, b):
        print(line)


#######
#
#   SYNTAX
#
#######


class SyntaxToken(TagToken):
    __attributes__ = ['id', 'text', 'head_id', 'rel']


class SyntaxMarkup(TagMarkup):
    __attributes__ = ['tokens']
    __annotations__ = {
        'tokens': [SyntaxToken]
    }

    @classmethod
    def from_tuples(cls, tuples):
        return cls([
            SyntaxToken(id, text, head_id, rel)
            for id, text, head_id, rel in tuples
        ])


def syntax_markup_deps(tokens):
    for token in tokens:
        id = int(token.id)
        head_id = int(token.head_id)
        # skip root=0, skip loop
        # ipymarkup crashes
        if head_id == 0 or head_id == id:
            continue

        rel = token.rel
        id = id - 1
        head_id = head_id - 1
        yield head_id, id, rel


def show_syntax_markup(markup):
    deps = syntax_markup_deps(markup.tokens)
    show_dep_markup(markup.words, deps)




import numpy as np


def pad_sequence(sequences, fill=0):
    # assert all sequences are 1d
    size = max(_.size for _ in sequences)
    array = np.full((len(sequences), size), fill)
    for index, sequence in enumerate(sequences):
        array[index, :sequence.size] = sequence
    return array




from os import getenv, environ
from os.path import exists, join
from random import seed, sample, randint, uniform
from itertools import islice as head
from subprocess import run

from tqdm.notebook import tqdm as log_progress

import torch
from torch import optim

from naeval.morph.datasets import load_dataset

from slovnet.s3 import S3
from slovnet.io import (
    format_jl,
    parse_jl,

    load_gz_lines,
    dump_gz_lines
)
from slovnet.board import (
    TensorBoard,
    LogBoard,
    MultiBoard
)
from slovnet.const import (
    TRAIN, TEST,
    PAD, CUDA0,
)

from slovnet.model.bert import (
    RuBERTConfig,
    BERTEmbedding,
    BERTEncoder,
    BERTMorphHead,
    BERTMorph
)
from slovnet.markup import MorphMarkup
from slovnet.vocab import BERTVocab, Vocab
from slovnet.encoders.bert import BERTMorphTrainEncoder
from slovnet.loss import masked_flatten_cross_entropy
from slovnet.batch import ProcessedBatch
from slovnet.score import (
    MorphScoreMeter,
    score_morph_batch
)
from slovnet.mask import (
    Masked,
    split_masked,
    pad_masked
)


DATA_DIR = 'data'
MODEL_DIR = 'model'
BERT_DIR = 'bert'
RAW_DIR = join(DATA_DIR, 'raw')

NEWS = join(DATA_DIR, 'news.jl.gz')
FICTION = join(DATA_DIR, 'fiction.jl.gz')
GRAMRU_DIR = join(RAW_DIR, 'GramEval2020-master')
GRAMRU_FILES = {
    NEWS: [
        'dataOpenTest/GramEval2020-RuEval2017-Lenta-news-dev.conllu',
        'dataTrain/MorphoRuEval2017-Lenta-train.conllu',
    ],
    FICTION: [
        'dataOpenTest/GramEval2020-SynTagRus-dev.conllu',
        'dataTrain/GramEval2020-SynTagRus-train-v2.conllu',
        'dataTrain/MorphoRuEval2017-JZ-gold.conllu'
    ],
}

S3_DIR = '03_bert_morph'
S3_NEWS = join(S3_DIR, NEWS)
S3_FICTION = join(S3_DIR, FICTION)

VOCAB = 'vocab.txt'
EMB = 'emb.pt'
ENCODER = 'encoder.pt'
MORPH = 'morph.pt'

BERT_VOCAB = join(BERT_DIR, VOCAB)
BERT_EMB = join(BERT_DIR, EMB)
BERT_ENCODER = join(BERT_DIR, ENCODER)

S3_RUBERT_DIR = '01_bert_news/rubert'
S3_MLM_DIR = '01_bert_news/model'
S3_BERT_VOCAB = join(S3_RUBERT_DIR, VOCAB)
S3_BERT_EMB = join(S3_MLM_DIR, EMB)
S3_BERT_ENCODER = join(S3_MLM_DIR, ENCODER)

TAGS_VOCAB = join(MODEL_DIR, 'tags_vocab.txt')
MODEL_ENCODER = join(MODEL_DIR, ENCODER)
MODEL_MORPH = join(MODEL_DIR, MORPH)

S3_TAGS_VOCAB = join(S3_DIR, TAGS_VOCAB)
S3_MODEL_ENCODER = join(S3_DIR, MODEL_ENCODER)
S3_MODEL_MORPH = join(S3_DIR, MODEL_MORPH)

BOARD_NAME = getenv('board_name', '03_bert_morph_02')
RUNS_DIR = 'runs'

TRAIN_BOARD = '01_train'
TEST_BOARD = '02_test'

SEED = int(getenv('seed', 1))
DEVICE = getenv('device', CUDA0)
BERT_LR = float(getenv('bert_lr', 0.0002))
LR = float(getenv('lr', 0.001))
LR_GAMMA = float(getenv('lr_gamma', 0.8))
EPOCHS = int(getenv('epochs', 5))


def process_batch(model, criterion, batch):
    input, target = batch

    pred = model(input.value)
    pred = pad_masked(pred, input.mask)
    mask = pad_masked(input.mask, input.mask)

    loss = criterion(pred, target.value, target.mask)

    pred = model.morph.decode(pred)
    pred = pred[mask]

    # unmask pred, target to fit in score batch
    target = target.value[target.mask]

    return ProcessedBatch(input, target, loss, pred)





from .record import Record
from .io import (
    load_lines,
    dump_lines
)
from .const import (
    B, I, O,

    UNK, PAD,
    CLS, SEP,
    MASK,
)
from .bio import format_bio


class Vocab(Record):
    __attributes__ = ['items']

    def __init__(self, items):
        self.items = items
        self.item_ids = {
            item: id
            for id, item in enumerate(self.items)
        }
        self.unk_id = self.item_ids.get(UNK)
        self.pad_id = self.item_ids.get(PAD)

    def encode(self, item):
        return self.item_ids.get(item, self.unk_id)

    def decode(self, id):
        return self.items[id]

    def __len__(self):
        return len(self.items)

    def __repr__(self):
        return '%s(items=[...])' % self.__class__.__name__

    def _repr_pretty_(self, printer, cycle):
        printer.text(repr(self))

    @classmethod
    def load(cls, path):
        items = list(load_lines(path))
        return cls(items)

    def dump(self, path):
        dump_lines(self.items, path)


class BERTVocab(Vocab):
    def __init__(self, items):
        super(BERTVocab, self).__init__(items)
        self.sep_id = self.item_ids[SEP]
        self.cls_id = self.item_ids[CLS]
        self.mask_id = self.item_ids[MASK]


class BIOTagsVocab(Vocab):
    def __init__(self, types):
        self.types = types

        items = [PAD, O]
        for type in types:
            for part in [B, I]:
                items.append(format_bio(part, type))

        super(BIOTagsVocab, self).__init__(items)




import torch

from slovnet.record import Record
from slovnet.chop import chop, chop_drop
from slovnet.shape import word_shape
from slovnet.batch import Batch

from .buffer import ShuffleBuffer
from .common import WordShapeInferEncoder


class TagTrainInput(Record):
    __attributes__ = ['word_id', 'shape_id']


class TagTrainEncoder:
    def __init__(self, words_vocab, shapes_vocab, tags_vocab,
                 seq_len=512, batch_size=8, shuffle_size=1):
        self.words_vocab = words_vocab
        self.shapes_vocab = shapes_vocab
        self.tags_vocab = tags_vocab

        self.seq_len = seq_len
        self.batch_size = batch_size

        self.shuffle = ShuffleBuffer(shuffle_size)

    def items(self, markups):
        for markup in markups:
            for token in markup.tokens:
                shape = word_shape(token.text)
                word = token.text.lower()
                yield (
                    self.words_vocab.encode(word),
                    self.shapes_vocab.encode(shape),
                    self.tags_vocab.encode(token.tag)
                )

    def batch(self, chunk):
        chunk = torch.tensor(chunk, dtype=torch.long)  # batch x seq x (word, shp, tag)
        word_id, shape_id, tag_id = chunk.unbind(-1)

        input = TagTrainInput(word_id, shape_id)
        return Batch(input, tag_id)

    def __call__(self, markups):
        items = self.items(markups)
        seqs = chop_drop(items, self.seq_len)
        seqs = self.shuffle(seqs)
        chunks = chop(seqs, self.batch_size)
        for chunk in chunks:
            yield self.batch(chunk)


class TagInferEncoder(WordShapeInferEncoder):
    pass





def format_conll_tag(pos, feats):
    if not feats:
        return pos

    feats = '|'.join(
        '%s=%s' % (_, feats[_])
        for _ in sorted(feats)
    )
    return '%s|%s' % (pos, feats)


def parse_conll_tag(tag):
    if '|' not in tag:
        return tag, {}

    pos, feats = tag.split('|', 1)
    feats = dict(
        _.split('=', 1)
        for _ in feats.split('|')
    )
    return pos, feats




import boto3

from .record import Record
from .const import (
    S3_KEY_ID,
    S3_KEY,
    S3_BUCKET,
    S3_ENDPOINT,
    S3_REGION
)


class S3(Record):
    __attributes__ = ['key_id', 'key', 'bucket', 'endpoint', 'region']

    def __init__(self, key_id=S3_KEY_ID, key=S3_KEY, bucket=S3_BUCKET,
                 endpoint=S3_ENDPOINT, region=S3_REGION):
        self.key_id = key_id
        self.key = key
        self.bucket = bucket
        self.endpoint = endpoint
        self.region = region

        self.client = boto3.client(
            's3',
            aws_access_key_id=key_id,
            aws_secret_access_key=key,
            region_name=region,
            endpoint_url=endpoint,
        )

    def upload(self, path, key):
        self.client.upload_file(path, self.bucket, key)

    def download(self, key, path):
        self.client.download_file(self.bucket, key, path)




import torch

from .record import Record
from .pad import pad_sequence


class Masked(Record):
    __attributes__ = ['value', 'mask']


def mask_like(input):
    return torch.ones_like(input, dtype=torch.bool)


def split_masked(input, mask):
    sizes = mask.sum(dim=-1).tolist()
    return input[mask].split(sizes)


def pad_masked(input, mask, fill=0):
    seqs = split_masked(input, mask)
    return pad_sequence(seqs, fill)


def fill_masked(input, mask, fill=0):
    return fill * mask + input * ~mask




from random import shuffle


class Buffer:
    def __init__(self, size):
        self.size = size
        self.reset()

    def append(self, item):
        self.buffer.append(item)

    def reset(self):
        self.buffer = []

    @property
    def is_full(self):
        return len(self.buffer) >= self.size

    def __call__(self, items):
        for item in items:
            self.append(item)
            if self.is_full:
                for item in self.flush():
                    yield item
        for item in self.flush():
            yield item


class ShuffleBuffer(Buffer):
    def flush(self):
        shuffle(self.buffer)
        for item in self.buffer:
            yield item
        self.reset()


class SortBuffer(Buffer):
    def __init__(self, size, key):
        self.key = key
        Buffer.__init__(self, size)

    def flush(self):
        self.buffer.sort(key=self.key)
        for item in self.buffer:
            yield item
        self.reset()




import pytest

from slovnet.token import tokenize
from slovnet.const import (
    B, I, O,
    PER, LOC
)
from slovnet.bio import (
    format_bio,

    io_spans,
    spans_io,

    bio_spans,
    spans_bio,

    bio_io,
    select_type_tags
)


T1, T2, T3, T4, T5 = tokenize('1 2 3 4 5')

B_PER = format_bio(B, PER)
I_PER = format_bio(I, PER)
B_LOC = format_bio(B, LOC)
I_LOC = format_bio(I, LOC)


TESTS = [
    [
        [T1, T2, T3],
        [O, O, O],
    ],
    [
        [],
        [],
    ]
]

IO_TESTS = [
    [
        [T1, T2, T3],
        [I_PER, O, O]
    ],
    [
        [T1, T2, T3],
        [I_PER, I_PER, O]
    ],
    [
        [T1, T2, T3],
        [I_PER, I_LOC, O]
    ],
    [
        [T1, T2],
        [I_PER, I_PER]
    ],
]

BIO_TESTS = [
    [
        [T1, T2, T3],
        [B_PER, O, O],
    ],
    [
        [T1, T2, T3],
        [B_PER, I_PER, O],
    ],
    [
        [T1, T2],
        [B_PER, I_PER],
    ],
    [
        [T1, T2, T3],
        [B_PER, B_LOC, O],
    ],
    [
        [T1, T2, T3],
        [B_PER, B_PER, O],
    ],
]

CONVERT_TESTS = [
    [
        [B_PER, I_PER],
        [I]
    ]
]


@pytest.mark.parametrize('test', TESTS + IO_TESTS)
def test_io(test):
    tokens, tags = test
    spans = io_spans(tokens, tags)
    guess = spans_io(tokens, spans)
    assert tags == list(guess)


@pytest.mark.parametrize('test', TESTS + BIO_TESTS)
def test_bio(test):
    tokens, tags = test
    spans = bio_spans(tokens, tags)
    guess = spans_bio(tokens, spans)
    assert tags == list(guess)


def test_convert():
    guess = bio_io([B_PER, I_PER, I_LOC])
    etalon = [I_PER, I_PER, I_LOC]
    assert etalon == list(guess)


def test_select():
    guess = select_type_tags([B_PER, I_LOC], PER)
    etalon = [B_PER, O]
    assert etalon == list(guess)




import json
import gzip


def load_lines(path, encoding='utf8'):
    with open(path, encoding=encoding) as file:
        for line in file:
            yield line.rstrip('\n')


def dump_lines(lines, path, encoding='utf8'):
    with open(path, 'w', encoding=encoding) as file:
        for line in lines:
            file.write(line + '\n')


def load_gz_lines(path, encoding='utf8'):
    with gzip.open(path) as file:
        for line in file:
            yield line.decode(encoding).rstrip()


def dump_gz_lines(lines, path):
    with gzip.open(path, 'wt') as file:
        for line in lines:
            file.write(line + '\n')


def load_json(path, encoding='utf8'):
    with open(path, encoding=encoding) as file:
        return json.load(file)


def format_jl(items):
    for item in items:
        yield json.dumps(item, ensure_ascii=False)


def parse_jl(lines):
    for line in lines:
        yield json.loads(line)




from os import getenv

import logging
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)-15s %(message)s'
)
log = logging.info

from aiohttp import web

from navec import Navec
from slovnet import Syntax


NAVEC = getenv('NAVEC', 'navec.tar')
PACK = getenv('PACK', 'pack.tar')
BATCH_SIZE = int(getenv('BATCH_SIZE', 8))

HOST = getenv('HOST', '0.0.0.0')
PORT = int(getenv('PORT', 8080))
MB = 1024 * 1024
MAX_SIZE = int(getenv('MAX_SIZE', 100 * MB))

log('Load navec: %r' % NAVEC)
navec = Navec.load(NAVEC)

log('Load pack: %r' % PACK)
log('Batch size: %r' % BATCH_SIZE)
syntax = Syntax.load(PACK)
syntax.navec(navec)


async def handle(request):
    chunk = await request.json()
    log('Post chunk size: %r' % len(chunk))
    markups = list(syntax.map(chunk))

    tokens = sum(len(_.tokens) for _ in markups)
    log('Infer tokens: %r', tokens)

    data = [_.as_json for _ in markups]
    return web.json_response(data)


log('Max size: %r' % (MAX_SIZE // MB))
app = web.Application(client_max_size=MAX_SIZE)
app.add_routes([web.post('/', handle)])

web.run_app(app, host=HOST, port=PORT)




from slovnet.record import Record
from slovnet.token import tokenize
from slovnet.markup import (
    BIOMarkup,
    MorphMarkup,
    SyntaxMarkup
)

from .mask import split_masked


class Infer(Record):
    __attributes__ = ['model', 'encoder', 'decoder']


######
#
#   TAG
#
#####


class TagDecoder(Record):
    __attributes__ = ['tags_vocab']

    def __call__(self, preds):
        for pred in preds:
            yield [self.tags_vocab.decode(_) for _ in pred]


def text_words(text):
    return [_.text for _ in tokenize(text)]


class NERInfer(Infer):
    def process(self, inputs):
        for input in inputs:
            pred = self.model(input.word_id, input.shape_id, input.pad_mask)
            yield from self.model.head.crf.decode(pred, ~input.pad_mask)

    def __call__(self, texts):
        items = [text_words(_) for _ in texts]
        inputs = self.encoder(items)
        preds = self.process(inputs)
        preds = self.decoder(preds)

        for text, item, pred in zip(texts, items, preds):
            tuples = zip(item, pred)
            markup = BIOMarkup.from_tuples(tuples)
            yield markup.to_span(text)


class MorphInfer(Infer):
    def process(self, inputs):
        for input in inputs:
            pred = self.model(input.word_id, input.shape_id, input.pad_mask)
            pred = self.model.head.decode(pred)
            yield from split_masked(pred, ~input.pad_mask)

    def __call__(self, items):
        inputs = self.encoder(items)
        preds = self.process(inputs)
        preds = self.decoder(preds)

        for item, pred in zip(items, preds):
            tuples = zip(item, pred)
            yield MorphMarkup.from_tuples(tuples)


########
#
#   SYNTAX
#
######


class SyntaxDecoder(Record):
    __attributes__ = ['rels_vocab']

    def __call__(self, preds):
        for pred in preds:
            head_ids, rel_ids = pred
            ids = [str(_ + 1) for _ in range(len(head_ids))]
            head_ids = [str(_) for _ in head_ids.tolist()]
            rels = [self.rels_vocab.decode(_) for _ in rel_ids]
            yield ids, head_ids, rels


class SyntaxInfer(Infer):
    def process(self, inputs):
        for input in inputs:
            pred = self.model(input.word_id, input.shape_id, input.pad_mask)
            mask = ~input.pad_mask

            head_id = self.model.head.decode(pred.head_id, mask)
            head_id = split_masked(head_id, mask)

            rel_id = self.model.rel.decode(pred.rel_id, mask)
            rel_id = split_masked(rel_id, mask)

            yield from zip(head_id, rel_id)

    def __call__(self, items):
        inputs = self.encoder(items)
        preds = self.process(inputs)
        preds = self.decoder(preds)

        for item, pred in zip(items, preds):
            ids, head_ids, rels = pred
            tuples = zip(ids, item, head_ids, rels)
            yield SyntaxMarkup.from_tuples(tuples)




from torch import nn

from .base import Module
from .crf import CRF
from .emb import WordShapeEmbedding
from .cnn import CNNEncoder


class NERHead(Module):
    def __init__(self, emb_dim, tags_num):
        super(NERHead, self).__init__()
        self.emb_dim = emb_dim
        self.tags_num = tags_num

        self.proj = nn.Linear(emb_dim, tags_num)
        self.crf = CRF(tags_num)

    def forward(self, input):
        return self.proj(input)


class MorphHead(Module):
    def __init__(self, emb_dim, tags_num):
        super(MorphHead, self).__init__()
        self.emb_dim = emb_dim
        self.tags_num = tags_num

        self.proj = nn.Linear(emb_dim, tags_num)

    def decode(self, pred):
        return pred.argmax(-1)

    def forward(self, input):
        return self.proj(input)


class TagEmbedding(WordShapeEmbedding):
    pass


class TagEncoder(CNNEncoder):
    pass


class Tag(Module):
    def __init__(self, emb, encoder, head):
        super(Tag, self).__init__()
        self.emb = emb
        self.encoder = encoder
        self.head = head

    def forward(self, word_id, shape_id, pad_mask=None):
        x = self.emb(word_id, shape_id)
        x = self.encoder(x, pad_mask)
        return self.head(x)


class NER(Tag):
    pass


class Morph(Tag):
    pass




import torch
from torch import nn

from .base import Module


class Embedding(nn.Embedding, Module):
    def __init__(self, vocab_size, dim, pad_id):
        super(Embedding, self).__init__(vocab_size, dim, pad_id)
        self.dim = dim


class PQEmbedding(Module):
    def __init__(self, indexes, codes):
        super(PQEmbedding, self).__init__()

        qdim, centroids, subdim = codes.shape
        self.subdim = subdim
        self.dim = qdim * subdim

        codes = codes.transpose(1, 0)  # for gather, centroids x qdim x subdim
        self.codes = nn.Parameter(codes, requires_grad=False)
        self.indexes = nn.Parameter(indexes, requires_grad=False)

    def extra_repr(self):
        return 'indexes=[...], codes=[...]'

    def forward(self, input):
        shape = input.shape
        input = input.flatten()  # reshape in return

        indexes = self.indexes[input]
        vectors, qdim = indexes.shape

        indexes = indexes.view(vectors, qdim, 1)
        indexes = indexes.expand(vectors, qdim, self.subdim)
        output = self.codes.gather(0, indexes.long())  # vectors x qdim x subdim
        return output.view(*shape, self.dim)


class NavecEmbedding(PQEmbedding):
    def __init__(self, navec):
        self.id = navec.meta.id
        super(NavecEmbedding, self).__init__(
            torch.from_numpy(navec.pq.indexes),
            torch.from_numpy(navec.pq.codes)
        )

    def extra_repr(self):
        return 'id=%r, indexes=[...], codes=[...]' % self.id


class WordShapeEmbedding(Module):
    def __init__(self, word, shape):
        super(WordShapeEmbedding, self).__init__()
        self.word = word
        self.shape = shape
        self.dim = word.dim + shape.dim

    def forward(self, word_id, shape_id):
        word = self.word(word_id)
        shape = self.shape(shape_id)
        return torch.cat([word, shape], dim=-1)




from .record import Record


class Batch(Record):
    __attributes__ = ['input', 'target']

    def processed(self, loss, pred):
        return ProcessedBatch(
            self.input, self.target,
            loss, pred
        )


class ProcessedBatch(Record):
    __attributes__ = ['input', 'target', 'loss', 'pred']




import torch
from torch import nn
from torch.nn import functional as F

from slovnet.record import Record
from slovnet.mask import fill_masked

from .base import Module
from .cnn import CNNEncoder
from .emb import WordShapeEmbedding


class FF(Module):
    def __init__(self, input_dim, hidden_dim, dropout):
        super(FF, self).__init__()
        self.proj = nn.Linear(input_dim, hidden_dim)
        self.relu = nn.ReLU()
        self.drop = nn.Dropout(dropout)

    def forward(self, input):
        x = self.proj(input)
        x = self.relu(x)
        return self.drop(x)


def append_root(input, root):
    batch_size, seq_len, emb_dim = input.shape
    root = root.repeat(batch_size).view(batch_size, 1, emb_dim)
    return torch.cat((root, input), dim=1)


def strip_root(input):
    input = input[:, 1:, :]
    return input.contiguous()


def append_root_mask(mask):
    return F.pad(mask, (1, 0), 'constant', True)


def matmul_mask(mask):
    # 1 1 1 0 0
    # ->
    # 1 1 1 0 0
    # 1 1 1 0 0
    # 1 1 1 0 0
    # 0 0 0 0 0
    # 0 0 0 0 0

    mask = mask.float()  # matmul not supported for bool
    mask = mask.unsqueeze(-2)  # batch x 1 x seq
    mask = mask.transpose(-2, -1).matmul(mask)
    return mask.bool()


class SyntaxHead(Module):
    def __init__(self, input_dim, hidden_dim, dropout=0.1):
        super(SyntaxHead, self).__init__()
        self.head = FF(input_dim, hidden_dim, dropout)
        self.tail = FF(input_dim, hidden_dim, dropout)

        self.root = nn.Parameter(torch.empty(input_dim))
        self.kernel = nn.Parameter(torch.empty(hidden_dim, hidden_dim))

        self.reset_parameters()

    def reset_parameters(self):
        nn.init.uniform_(self.root)
        nn.init.eye_(self.kernel)

    def decode(self, pred, mask):
        # multiple roots
        # loops, nonprojective
        # ~10% sents

        mask = append_root_mask(mask)
        mask = matmul_mask(mask)
        mask = strip_root(mask)

        pred = fill_masked(pred, ~mask, pred.min())
        return pred.argmax(-1)

    def forward(self, input):
        input = append_root(input, self.root)
        head = self.head(input)
        tail = self.tail(input)

        x = head.matmul(self.kernel)
        x = x.matmul(tail.transpose(-2, -1))
        return strip_root(x)


def gather_head(input, root, index):
    batch_size, seq_len, emb_dim = input.shape
    input = append_root(input, root)  # batch x seq + 1 x emb

    # for root select root
    zero = torch.zeros(batch_size, 1, dtype=torch.long, device=input.device)
    index = torch.cat((zero, index), dim=-1)  # batch x seq + 1 x emb

    # prep for gather
    index = index.view(batch_size, seq_len + 1, 1)
    index = index.expand(batch_size, seq_len + 1, emb_dim)

    input = torch.gather(input, dim=-2, index=index)
    return strip_root(input)  # batch x seq x emb


class SyntaxRel(Module):
    def __init__(self, input_dim, hidden_dim, rel_dim, dropout=0.1):
        super(SyntaxRel, self).__init__()
        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.rel_dim = rel_dim

        self.head = FF(input_dim, hidden_dim, dropout)
        self.tail = FF(input_dim, hidden_dim, dropout)

        self.root = nn.Parameter(torch.empty(input_dim))
        self.kernel = nn.Parameter(torch.empty(hidden_dim, hidden_dim * rel_dim))

        self.reset_parameters()

    def reset_parameters(self):
        nn.init.uniform_(self.root)
        nn.init.xavier_uniform_(self.kernel)

    def decode(self, pred, mask):
        mask = mask.unsqueeze(-1)  # batch x seq x 1
        mask = mask.expand_as(pred)

        pred = fill_masked(pred, ~mask, pred.min())
        return pred.argmax(-1)

    def forward(self, input, head_id):
        head = self.head(gather_head(input, self.root, head_id))
        tail = self.tail(input)

        batch_size, seq_len, _ = input.shape
        x = head.matmul(self.kernel)  # batch x seq x hidden * rel
        x = x.view(batch_size, seq_len, self.rel_dim, self.hidden_dim)
        x = x.matmul(tail.view(batch_size, seq_len, self.hidden_dim, 1))
        return x.view(batch_size, seq_len, self.rel_dim)


class SyntaxPred(Record):
    __attributes__ = ['head_id', 'rel_id']


class SyntaxEmbedding(WordShapeEmbedding):
    pass


class SyntaxEncoder(CNNEncoder):
    pass


class Syntax(Module):
    def __init__(self, emb, encoder, head, rel):
        super(Syntax, self).__init__()
        self.emb = emb
        self.encoder = encoder
        self.head = head
        self.rel = rel

    def forward(self, word_id, shape_id, pad_mask, target_head_id=None):
        x = self.emb(word_id, shape_id)
        x = self.encoder(x, pad_mask)

        head_id = self.head(x)
        if target_head_id is None:
            target_head_id = self.head.decode(head_id, ~pad_mask)

        return SyntaxPred(
            head_id=head_id,
            rel_id=self.rel(x, target_head_id)
        )




from torch import nn

from .state import StateMixin
from .exec import ExecMixin


class Module(nn.Module, StateMixin, ExecMixin):
    @property
    def device(self):
        for parameter in self.parameters():
            return parameter.device




import torch

from slovnet.record import Record
from slovnet.pad import pad_sequence
from slovnet.chop import chop
from slovnet.shape import word_shape


class WordShapeInferInput(Record):
    __attributes__ = ['word_id', 'shape_id', 'pad_mask']


class WordShapeInferEncoder:
    def __init__(self, words_vocab, shapes_vocab,
                 batch_size=8):
        self.words_vocab = words_vocab
        self.shapes_vocab = shapes_vocab

        self.batch_size = batch_size

    def item(self, words):
        word_ids, shape_ids = [], []
        for word in words:
            shape = word_shape(word)
            word_id = self.words_vocab.encode(word.lower())
            shape_id = self.shapes_vocab.encode(shape)
            word_ids.append(word_id)
            shape_ids.append(shape_id)
        return word_ids, shape_ids

    def input(self, items):
        word_id, shape_id = [], []
        for word_ids, shape_ids in items:
            word_id.append(torch.tensor(word_ids, dtype=torch.long))
            shape_id.append(torch.tensor(shape_ids, dtype=torch.long))
        word_id = pad_sequence(word_id, self.words_vocab.pad_id)
        shape_id = pad_sequence(shape_id, self.shapes_vocab.pad_id)
        pad_mask = word_id == self.words_vocab.pad_id
        return WordShapeInferInput(word_id, shape_id, pad_mask)

    def __call__(self, items):
        items = (self.item(_) for _ in items)
        chunks = chop(items, self.batch_size)
        for chunk in chunks:
            yield self.input(chunk)




from torch.nn import functional as F

from .mask import fill_masked


def flatten_cross_entropy(pred, target, ignore_id=None):
    target = target.flatten()
    pred = pred.view(len(target), -1)
    return F.cross_entropy(pred, target, ignore_index=ignore_id)


def masked_flatten_cross_entropy(pred, target, mask, ignore_id=-100):
    target = fill_masked(target, ~mask, ignore_id)
    return flatten_cross_entropy(pred, target, ignore_id)




import torch

from slovnet.record import Record
from slovnet.pad import pad_sequence
from slovnet.chop import chop
from slovnet.shape import word_shape
from slovnet.batch import Batch

from .buffer import SortBuffer
from .common import WordShapeInferEncoder


ROOT_ID = '0'


class SyntaxTrainItem(Record):
    __attributes__ = ['word_ids', 'shape_ids', 'head_ids', 'rel_ids']


class SyntaxInput(Record):
    __attributes__ = ['word_id', 'shape_id', 'pad_mask']


class SyntaxTarget(Record):
    __attributes__ = ['head_id', 'rel_id', 'mask']


class SyntaxTrainEncoder:
    def __init__(self, words_vocab, shapes_vocab, rels_vocab,
                 batch_size=8, sort_size=1):
        self.words_vocab = words_vocab
        self.shapes_vocab = shapes_vocab
        self.rels_vocab = rels_vocab

        self.batch_size = batch_size
        self.sort = SortBuffer(sort_size, key=lambda _: len(_.tokens))

    def item(self, markup):
        word_ids, shape_ids, head_ids, rel_ids = [], [], [], []
        ids = {ROOT_ID: 0}

        for index, token in enumerate(markup.tokens, 1):
            ids[token.id] = index
            head_ids.append(token.head_id)

            rel_id = self.rels_vocab.encode(token.rel)
            rel_ids.append(rel_id)

            shape = word_shape(token.text)
            shape_id = self.shapes_vocab.encode(shape)
            shape_ids.append(shape_id)

            word = token.text.lower()
            word_id = self.words_vocab.encode(word)
            word_ids.append(word_id)

        head_ids = [ids[_] for _ in head_ids]
        return SyntaxTrainItem(word_ids, shape_ids, head_ids, rel_ids)

    def batch(self, chunk):
        word_id, shape_id, head_id, rel_id = [], [], [], []
        for item in chunk:
            word_id.append(torch.tensor(item.word_ids, dtype=torch.long))
            shape_id.append(torch.tensor(item.shape_ids, dtype=torch.long))
            head_id.append(torch.tensor(item.head_ids, dtype=torch.long))
            rel_id.append(torch.tensor(item.rel_ids, dtype=torch.long))

        word_id = pad_sequence(word_id, fill=self.words_vocab.pad_id)
        shape_id = pad_sequence(shape_id, fill=self.shapes_vocab.pad_id)
        pad_mask = word_id == self.words_vocab.pad_id
        input = SyntaxInput(word_id, shape_id, pad_mask)

        head_id = pad_sequence(head_id)
        rel_id = pad_sequence(rel_id, fill=self.rels_vocab.pad_id)
        mask = rel_id != self.rels_vocab.pad_id
        target = SyntaxTarget(head_id, rel_id, mask)

        return Batch(input, target)

    def __call__(self, markups):
        markups = self.sort(markups)
        items = (self.item(_) for _ in markups)
        chunks = chop(items, self.batch_size)
        for chunk in chunks:
            yield self.batch(chunk)


class SyntaxInferEncoder(WordShapeInferEncoder):
    pass




import pytest

from os.path import join, dirname, basename, exists
from os import makedirs
from urllib.request import urlopen
from shutil import copyfileobj

from navec import Navec
from slovnet import NER, Morph, Syntax


DATA_DIR = join(dirname(__file__), '../data/test')


def download(url, dir=DATA_DIR):
    path = join(dir, basename(url))
    if exists(path):
        return path

    if not exists(dir):
        makedirs(dir)

    with urlopen(url) as source:
        with open(path, 'wb') as target:
            copyfileobj(source, target)

    return path


@pytest.fixture(scope='module')
def navec():
    path = download('https://storage.yandexcloud.net/natasha-navec/packs/navec_news_v1_1B_250K_300d_100q.tar')
    return Navec.load(path)


@pytest.fixture(scope='module')
def ner(navec):
    path = download('https://storage.yandexcloud.net/natasha-slovnet/packs/slovnet_ner_news_v1.tar')
    return NER.load(path).navec(navec)


@pytest.fixture(scope='module')
def morph(navec):
    path = download('https://storage.yandexcloud.net/natasha-slovnet/packs/slovnet_morph_news_v1.tar')
    return Morph.load(path).navec(navec)


@pytest.fixture(scope='module')
def syntax(navec):
    path = download('https://storage.yandexcloud.net/natasha-slovnet/packs/slovnet_syntax_news_v1.tar')
    return Syntax.load(path).navec(navec)


def test_ner(ner):
    text = 'На них удержали лидерство действующие руководители и партии — Денис Пушилин и «Донецкая республика» в ДНР и Леонид Пасечник с движением «Мир Луганщине» в ЛНР.'

    markup = ner(text)

    pred = []
    for span in markup.spans:
        chunk = markup.text[span.start:span.stop]
        pred.append([span.type, chunk])

    assert pred == [
        ['PER', 'Денис Пушилин'],
        ['ORG', 'Донецкая республика'],
        ['LOC', 'ДНР'],
        ['PER', 'Леонид Пасечник'],
        ['ORG', 'Мир Луганщине'],
        ['LOC', 'ЛНР']
    ]


def test_morph(morph):
    words = ['Об', 'этом', 'говорится', 'в', 'документе', ',', 'опубликованном', 'в', 'официальном', 'журнале', 'Евросоюза', '.']

    markup = morph(words)

    pred = [
        [_.text, _.tag]
        for _ in markup.tokens
    ]
    assert pred == [
        ['Об', 'ADP'],
        ['этом', 'PRON|Animacy=Inan|Case=Loc|Gender=Neut|Number=Sing'],
        ['говорится', 'VERB|Aspect=Imp|Mood=Ind|Number=Sing|Person=3|Tense=Pres|VerbForm=Fin|Voice=Pass'],
        ['в', 'ADP'],
        ['документе', 'NOUN|Animacy=Inan|Case=Loc|Gender=Masc|Number=Sing'],
        [',', 'PUNCT'],
        ['опубликованном', 'VERB|Aspect=Perf|Case=Loc|Gender=Masc|Number=Sing|Tense=Past|VerbForm=Part|Voice=Pass'],
        ['в', 'ADP'],
        ['официальном', 'ADJ|Case=Loc|Degree=Pos|Gender=Masc|Number=Sing'],
        ['журнале', 'NOUN|Animacy=Inan|Case=Loc|Gender=Masc|Number=Sing'],
        ['Евросоюза', 'PROPN|Animacy=Inan|Case=Gen|Gender=Masc|Number=Sing'],
        ['.', 'PUNCT']
    ]


def test_syntax(syntax):
    words = ['Опубликованы', 'новые', 'данные', 'по', 'заражению', 'коронавирусом', 'в', 'Москве']

    markup = syntax(words)

    ids = {_.id: _ for _ in markup.tokens}
    pred = []
    for token in markup.tokens:
        head = ids.get(token.head_id)
        if head:
            pred.append([token.text, head.rel, head.text])
        else:
            pred.append(token.text)

    assert pred == [
        'Опубликованы',
        ['новые', 'nsubj:pass', 'данные'],
        ['данные', 'root', 'Опубликованы'],
        ['по', 'nmod', 'заражению'],
        ['заражению', 'nsubj:pass', 'данные'],
        ['коронавирусом', 'nmod', 'заражению'],
        ['в', 'obl', 'Москве'],
        ['Москве', 'nmod', 'коронавирусом']
    ]




from itertools import groupby

from slovnet.record import Record
from slovnet.token import tokenize
from slovnet.bert import bert_subs
from slovnet.chop import chop_weighted
from slovnet.mask import pad_masked, split_masked
from slovnet.markup import (
    BIOMarkup,
    MorphMarkup,
    SyntaxMarkup
)

from .base import Infer
from .tag import TagDecoder
from .syntax import SyntaxDecoder


##########
#
#   SEGMENT
#
#########


class SubsToken(Record):
    __attributes__ = ['text', 'subs']


class BERTInferItem(Record):
    __attributes__ = ['id', 'tokens', 'pred']

    def __init__(self, id, tokens, pred=None):
        self.id = id
        self.tokens = tokens
        self.pred = pred

    @property
    def words(self):
        return [_.text for _ in self.tokens]


def substoken(text, vocab):
    subs = bert_subs(text, vocab)
    return SubsToken(text, subs)


def text_items(texts, vocab):
    for id, text in enumerate(texts):
        tokens = [
            substoken(_.text, vocab)
            for _ in tokenize(text)
        ]
        yield BERTInferItem(id, tokens)


def word_items(items, vocab):
    for id, words in enumerate(items):
        tokens = [
            substoken(_, vocab)
            for _ in words
        ]
        yield BERTInferItem(id, tokens)


def segment_items(items, seq_len):
    for item in items:
        chunks = chop_weighted(
            item.tokens,
            seq_len,
            weight=lambda _: len(_.subs)
        )
        for chunk in chunks:
            yield BERTInferItem(item.id, chunk)


def flatten(seqs):
    return [
        item
        for seq in seqs
        for item in seq
    ]


def join_items(items):
    for id, group in groupby(items, key=lambda _: _.id):
        group = list(group)
        tokens = flatten(_.tokens for _ in group)
        pred = flatten(_.pred for _ in group)
        yield BERTInferItem(id, tokens, pred)


#######
#
#  DECODE
#
#######


class BERTTagDecoder(TagDecoder):
    pass


class BERTSyntaxDecoder(SyntaxDecoder):
    pass


#######
#
#   INFER
#
######


class BERTNERInfer(Infer):
    def process(self, inputs):
        for input in inputs:
            input = input.to(self.model.device)
            pred = self.model(input.word_id, input.pad_mask)
            pred = pad_masked(pred, input.word_mask)
            mask = pad_masked(input.word_mask, input.word_mask)
            yield from self.model.head.crf.decode(pred, mask)

    def __call__(self, chunk):
        items = text_items(chunk, self.encoder.words_vocab)
        # consider <cls>, <sep> spec tokens
        items = list(segment_items(items, self.encoder.seq_len - 2))
        inputs = self.encoder(items)
        preds = self.process(inputs)
        preds = self.decoder(preds)

        for item, pred in zip(items, preds):
            item.pred = pred

        items = join_items(items)

        for text, item in zip(chunk, items):
            tuples = zip(item.words, item.pred)
            markup = BIOMarkup.from_tuples(tuples)
            yield markup.to_span(text)


class BERTMorphInfer(Infer):
    def process(self, inputs):
        for input in inputs:
            input = input.to(self.model.device)
            pred = self.model(input.word_id, input.pad_mask)
            pred = self.model.head.decode(pred)
            yield from split_masked(pred, input.word_mask)

    def __call__(self, chunk):
        items = word_items(chunk, self.encoder.words_vocab)
        items = list(segment_items(items, self.encoder.seq_len - 2))
        inputs = self.encoder(items)
        preds = self.process(inputs)
        preds = self.decoder(preds)

        for item, pred in zip(items, preds):
            item.pred = pred

        items = join_items(items)
        for item in items:
            tuples = zip(item.words, item.pred)
            yield MorphMarkup.from_tuples(tuples)


class BERTSyntaxInfer(Infer):
    def process(self, inputs):
        for input in inputs:
            input = input.to(self.model.device)
            pred = self.model(input.word_id, input.word_mask, input.pad_mask)
            head_id = self.model.head.decode(pred.head_id)
            rel_id = self.model.rel.decode(pred.rel_id)
            yield from zip(head_id, rel_id)

    def __call__(self, chunk):
        items = list(word_items(chunk, self.encoder.words_vocab))
        inputs = self.encoder(items)
        preds = self.process(inputs)
        preds = self.decoder(preds)

        for item, pred in zip(items, preds):
            item.pred = pred

        items = join_items(items)
        for item in items:
            ids, head_ids, rels = item.pred
            tuples = zip(ids, item.words, head_ids, rels)
            yield SyntaxMarkup.from_tuples(tuples)




import torch

from slovnet.visitor import Visitor
from slovnet.exec import model as exec


class ExecVisitor(Visitor):
    def visit_Parameter(self, item):
        return self.visit(item.data)

    def visit_Tensor(self, item):
        array = item.detach().numpy()
        return exec.Weight(
            array.shape,
            array.dtype.name,
            array
        )

    def visit_Linear(self, item):
        # in torch linear is xA^T + b
        weight = item.weight.transpose(1, 0)
        return exec.Linear(
            self.visit(weight),
            self.visit(item.bias)
        )

    def visit_Conv1d(self, item):
        padding, = item.padding  # tuple -> int
        return exec.Conv1d(
            self.visit(item.weight),
            self.visit(item.bias),
            padding
        )

    def visit_ReLU(self, item):
        return exec.ReLU()

    def visit_BatchNorm1d(self, item):
        running_std = torch.sqrt(item.running_var + item.eps)
        return exec.BatchNorm1d(
            self.visit(item.weight),
            self.visit(item.bias),
            self.visit(item.running_mean),
            self.visit(running_std),
        )

    def visit_Embedding(self, item):
        return exec.Embedding(
            self.visit(item.weight)
        )

    def visit_NavecEmbedding(self, item):
        # recover initial qdim x centroids x chunk
        codes = item.codes.transpose(1, 0)
        return exec.NavecEmbedding(
            item.id,
            self.visit(item.indexes),
            self.visit(codes)
        )

    def visit_WordShapeEmbedding(self, item):
        return exec.WordShapeEmbedding(
            self.visit(item.word),
            self.visit(item.shape)
        )

    def visit_CNNEncoderLayer(self, item):
        return exec.CNNEncoderLayer(
            self.visit(item.conv),
            self.visit(item.relu),
            self.visit(item.norm)
        )

    def visit_CNNEncoder(self, item):
        return exec.CNNEncoder([
            self.visit(_)
            for _ in item.layers
        ])

    def visit_NERHead(self, item):
        return exec.NERHead(
            self.visit(item.proj),
            self.visit(item.crf)
        )

    def visit_MorphHead(self, item):
        return exec.MorphHead(
            self.visit(item.proj)
        )

    def visit_Tag(self, item):
        from slovnet.model.tag import NERHead, MorphHead

        cls = type(item.head)
        if cls is NERHead:
            Tag = exec.NER
        elif cls is MorphHead:
            Tag = exec.Morph

        return Tag(
            self.visit(item.emb),
            self.visit(item.encoder),
            self.visit(item.head)
        )

    def visit_FF(self, item):
        return exec.FF(
            self.visit(item.proj),
            self.visit(item.relu)
        )

    def visit_SyntaxHead(self, item):
        return exec.SyntaxHead(
            self.visit(item.head),
            self.visit(item.tail),
            self.visit(item.root),
            self.visit(item.kernel)
        )

    def visit_SyntaxRel(self, item):
        return exec.SyntaxRel(
            self.visit(item.head),
            self.visit(item.tail),
            self.visit(item.root),
            self.visit(item.kernel)
        )

    def visit_Syntax(self, item):
        return exec.Syntax(
            self.visit(item.emb),
            self.visit(item.encoder),
            self.visit(item.head),
            self.visit(item.rel)
        )

    def visit_CRF(self, item):
        return exec.CRF(
            self.visit(item.transitions)
        )


class ExecMixin:
    # super stange error if as_exec property
    # torch Module does some magic
    def to_exec(self):
        visitor = ExecVisitor()
        return visitor(self)




import torch
from torch import nn
from torch.nn import functional as F

from slovnet.record import Record
from slovnet.mask import pad_masked

from .base import Module
from .tag import (
    NERHead,
    MorphHead
)
from .syntax import (
    SyntaxHead,
    SyntaxRel,
    SyntaxPred
)


class BERTConfig(Record):
    __attributes__ = [
        'vocab_size', 'seq_len', 'emb_dim',
        'layers_num', 'heads_num', 'hidden_dim',
        'dropout', 'norm_eps'
    ]


class RuBERTConfig(BERTConfig):
    def __init__(self,
                 vocab_size=50106,
                 seq_len=512,
                 emb_dim=768,
                 layers_num=12,
                 heads_num=12,
                 hidden_dim=3072,
                 dropout=0.1,
                 norm_eps=1e-12):
        super(RuBERTConfig, self).__init__(
            vocab_size, seq_len, emb_dim,
            layers_num, heads_num, hidden_dim,
            dropout, norm_eps
        )


class BERTEmbedding(Module):
    def __init__(self, vocab_size, seq_len, emb_dim, dropout=0.1, norm_eps=1e-12):
        super(BERTEmbedding, self).__init__()
        self.word = nn.Embedding(vocab_size, emb_dim)
        self.position = nn.Embedding(seq_len, emb_dim)
        self.norm = nn.LayerNorm(emb_dim, eps=norm_eps)
        self.drop = nn.Dropout(dropout)

    @classmethod
    def from_config(cls, config):
        return cls(
            config.vocab_size, config.seq_len, config.emb_dim,
            config.dropout, config.norm_eps
        )

    def forward(self, input):
        batch_size, seq_len = input.shape
        position = torch.arange(seq_len).expand_as(input).to(input.device)

        emb = self.word(input) + self.position(position)
        emb = self.norm(emb)
        return self.drop(emb)


def BERTLayer(emb_dim, heads_num, hidden_dim, dropout=0.1, norm_eps=1e-12):
    layer = nn.TransformerEncoderLayer(
        d_model=emb_dim,
        nhead=heads_num,
        dim_feedforward=hidden_dim,
        dropout=dropout,
        activation='gelu'
    )
    layer.norm1.eps = norm_eps
    layer.norm2.eps = norm_eps
    return layer


class BERTEncoder(Module):
    def __init__(self, layers_num, emb_dim, heads_num, hidden_dim,
                 dropout=0.1, norm_eps=1e-12):
        super(BERTEncoder, self).__init__()
        self.layers = nn.ModuleList([
            BERTLayer(
                emb_dim, heads_num, hidden_dim,
                dropout, norm_eps
            )
            for _ in range(layers_num)
        ])

    @classmethod
    def from_config(cls, config):
        return cls(
            config.layers_num, config.emb_dim, config.heads_num, config.hidden_dim,
            config.dropout, config.norm_eps
        )

    def forward(self, input, pad_mask=None):
        input = input.transpose(0, 1)  # torch expects seq x batch x emb
        for layer in self.layers:
            input = layer(input, src_key_padding_mask=pad_mask)
        return input.transpose(0, 1)  # restore


#########
#
#   MLM
#
#########


class BERTMLMHead(Module):
    def __init__(self, emb_dim, vocab_size, norm_eps=1e-12):
        super(BERTMLMHead, self).__init__()
        self.linear1 = nn.Linear(emb_dim, emb_dim)
        self.norm = nn.LayerNorm(emb_dim, eps=norm_eps)
        self.linear2 = nn.Linear(emb_dim, vocab_size)

    def forward(self, input):
        x = self.linear1(input)
        x = F.gelu(x)
        x = self.norm(x)
        return self.linear2(x)


class BERTMLM(Module):
    def __init__(self, emb, encoder, head):
        super(BERTMLM, self).__init__()
        self.emb = emb
        self.encoder = encoder
        self.head = head

    def forward(self, input):
        x = self.emb(input)
        x = self.encoder(x)
        return self.head(x)


#########
#
#   TAG
#
######


class BERTNERHead(NERHead):
    pass


class BERTMorphHead(MorphHead):
    pass


class BERTTag(Module):
    def __init__(self, emb, encoder, head):
        super(BERTTag, self).__init__()
        self.emb = emb
        self.encoder = encoder
        self.head = head

    def forward(self, input, pad_mask=None):
        x = self.emb(input)
        x = self.encoder(x, pad_mask)
        return self.head(x)


class BERTNER(BERTTag):
    pass


class BERTMorph(BERTTag):
    pass


#######
#
#   SYNTAX
#
#######


class BERTSyntaxHead(SyntaxHead):
    pass


class BERTSyntaxRel(SyntaxRel):
    pass


class BERTSyntax(Module):
    def __init__(self, emb, encoder, head, rel):
        super(BERTSyntax, self).__init__()
        self.emb = emb
        self.encoder = encoder
        self.head = head
        self.rel = rel

    def forward(self, input, word_mask, pad_mask,
                target_mask, target_head_id=None):
        x = self.emb(input)
        x = self.encoder(x, pad_mask)
        x = pad_masked(x, word_mask)

        head_id = self.head(x)
        if target_head_id is None:
            target_head_id = self.head.decode(head_id, target_mask)

        return SyntaxPred(
            head_id=head_id,
            rel_id=self.rel(x, target_head_id)
        )




import re

RU = 'RU'
EN = 'EN'
NUM = 'NUM'
PUNCT = 'PUNCT'
OTHER = 'OTHER'

PUNCTS = (
    '!#$%&()[]\\/*+,.:;<=>?@^_{|}~'  # string.punctuation
    '-‐−‒⁃–—―'  # https://habr.com/ru/post/20588/
    '`"\'«»„“ʼʻ”'
    '№…'
)
TYPE = re.compile(
    r'''
    (?P<RU>[а-яё]+)
    |(?P<EN>[a-z]+)
    |(?P<NUM>[+-]?\d+)
    |(?P<PUNCT>[%s]+)
    ''' % re.escape(PUNCTS),
    re.X | re.IGNORECASE
)

X = 'X'
x = 'x'
XX = 'XX'
xx = 'xx'
Xx = 'Xx'
Xx_Xx = 'Xx-Xx'


def is_title(word):
    return len(word) > 1 and word[0].isupper() and word[1:].islower()


def is_dash_title(word):
    if '-' in word:
        left, right = word.split('-', 1)
        return is_title(left) and is_title(right)


def word_outline(word):
    if len(word) == 1:
        if word.isupper():
            return X
        else:
            return x
    else:
        if word.isupper():
            return XX
        elif word.islower():
            return xx
        elif is_title(word):
            return Xx
        elif is_dash_title(word):
            return Xx_Xx
        else:
            return OTHER


def word_type(word):
    # СИЗО-6 -> RU
    # 2011-2020 -> NUM
    match = TYPE.match(word)
    if match:
        return match.lastgroup
    return OTHER


def format_shape(type, value):
    return '%s_%s' % (type, value)


def word_shape(word):
    type = word_type(word)
    if type in (RU, EN):
        return format_shape(type, word_outline(word))
    elif type == PUNCT:
        if len(word) > 1 or word not in PUNCTS:
            # ..., ?!, ****
            word = OTHER
        return format_shape(PUNCT, word)
    elif type in (NUM, OTHER):
        return type


OUTLINES = [X, x, XX, xx, Xx, Xx_Xx, OTHER]
SHAPES = (
    [format_shape(RU, _) for _ in OUTLINES]
    + [format_shape(EN, _) for _ in OUTLINES]
    + [format_shape(PUNCT, _) for _ in PUNCTS]
    + [format_shape(PUNCT, OTHER), NUM, OTHER]
)




import sys
from datetime import datetime


def log(format, *args):
    message = format % args
    now = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
    print(
        '[%s] %s' % (now, message),
        file=sys.stderr
    )





def split_masked(input, mask):
    sizes = mask.sum(-1)
    for index, size in enumerate(sizes):
        yield input[index, :size]


def fill_masked(input, mask, fill=0):
    return fill * mask + input * ~mask




from slovnet.mask import split_masked
from slovnet.token import tokenize
from slovnet.markup import (
    BIOMarkup,
    MorphMarkup
)

from .base import Infer


class TagDecoder:
    def __init__(self, tags_vocab):
        self.tags_vocab = tags_vocab

    def __call__(self, preds):
        for pred in preds:
            yield [self.tags_vocab.decode(_) for _ in pred]


def text_words(text):
    return [_.text for _ in tokenize(text)]


class NERInfer(Infer):
    def process(self, inputs):
        for input in inputs:
            input = input.to(self.model.device)
            pred = self.model(input.word_id, input.shape_id, input.pad_mask)
            yield from self.model.ner.crf.decode(pred, ~input.pad_mask)

    def __call__(self, texts):
        items = [text_words(_) for _ in texts]
        inputs = self.encoder(items)
        preds = self.process(inputs)
        preds = self.decoder(preds)

        for text, item, pred in zip(texts, items, preds):
            tuples = zip(item, pred)
            markup = BIOMarkup.from_tuples(tuples)
            yield markup.to_span(text)


class MorphInfer(Infer):
    def process(self, inputs):
        for input in inputs:
            input = input.to(self.model.device)
            pred = self.model(input.word_id, input.shape_id, input.pad_mask)
            pred = self.model.morph.decode(pred)
            yield from split_masked(pred, ~input.pad_mask)

    def __call__(self, items):
        inputs = self.encoder(items)
        preds = self.process(inputs)
        preds = self.decoder(preds)

        for item, pred in zip(items, preds):
            tuples = zip(item, pred)
            yield MorphMarkup.from_tuples(tuples)




from collections import OrderedDict


def parse_annotation(annotation):
    type = annotation or str

    repeatable = False
    if isinstance(annotation, list):  # [Fact]
        repeatable = True
        type = annotation[0]

    is_record = issubclass(type, Record)

    return type, repeatable, is_record


class Record(object):
    __attributes__ = []
    __annotations__ = {}

    def __init__(self, *args, **kwargs):
        for key, value in zip(self.__attributes__, args):
            self.__dict__[key] = value
        self.__dict__.update(kwargs)

    def __eq__(self, other):
        return (
            type(self) == type(other)
            and all(
                (getattr(self, _) == getattr(other, _))
                for _ in self.__attributes__
            )
        )

    def __ne__(self, other):
        return not self == other

    def __iter__(self):
        return (getattr(self, _) for _ in self.__attributes__)

    def __hash__(self):
        return hash(tuple(self))

    def __repr__(self):
        name = self.__class__.__name__
        args = ', '.join(
            '{key}={value!r}'.format(
                key=_,
                value=getattr(self, _)
            )
            for _ in self.__attributes__
        )
        return '{name}({args})'.format(
            name=name,
            args=args
        )

    def _repr_pretty_(self, printer, cycle):
        name = self.__class__.__name__
        if cycle:
            printer.text('{name}(...)'.format(name=name))
        else:
            printer.text('{name}('.format(name=name))
            keys = self.__attributes__
            size = len(keys)
            if size:
                with printer.indent(4):
                    printer.break_()
                    for index, key in enumerate(keys):
                        printer.text(key + '=')
                        value = getattr(self, key)
                        printer.pretty(value)
                        if index < size - 1:
                            printer.text(',')
                            printer.break_()
                printer.break_()
            printer.text(')')

    @property
    def as_json(self):
        data = OrderedDict()
        for key in self.__attributes__:
            annotation = self.__annotations__.get(key)
            _, repeatable, is_record = parse_annotation(annotation)

            value = getattr(self, key)
            if value is None:
                continue

            if repeatable and is_record:
                value = [_.as_json for _ in value]
            elif is_record:
                value = value.as_json

            data[key] = value
        return data

    @classmethod
    def from_json(cls, data):
        args = []
        for key in cls.__attributes__:
            annotation = cls.__annotations__.get(key)
            type, repeatable, is_record = parse_annotation(annotation)
            value = data.get(key)
            if value is None and repeatable:
                value = []
            elif value is not None:
                if repeatable and is_record:
                    value = [type.from_json(_) for _ in value]
                elif is_record:
                    value = type.from_json(value)
            args.append(value)
        return cls(*args)

    def to(self, device):
        cls = type(self)
        args = (_.to(device) for _ in self)
        return cls(*args)

    def copy(self):
        return type(self)(*self)

    def replace(self, **kwargs):
        other = self.copy()
        for key, value in kwargs.items():
            setattr(other, key, value)
        return other




import torch
from torch import nn

from slovnet.mask import mask_like

from .base import Module


class CRF(Module):
    # https://github.com/kmkurn/pytorch-crf/blob/master/torchcrf/__init__.py

    def __init__(self, tags_num):
        super(CRF, self).__init__()
        self.tags_num = tags_num
        self.transitions = nn.Parameter(torch.empty(tags_num, tags_num))
        self.reset_parameters()

    def reset_parameters(self):
        nn.init.uniform_(self.transitions, -0.1, 0.1)

    def extra_repr(self):
        return 'tags_num=%d' % self.tags_num

    def forward(self, emissions, tags, mask=None):
        if mask is None:
            mask = mask_like(tags)

        emissions = emissions.transpose(1, 0)  # seq x batch x tags
        tags = tags.transpose(1, 0)  # seq x batch
        mask = mask.transpose(1, 0)

        log_likelihood = (
            self.score(emissions, tags, mask)
            - self.normalization(emissions, mask)
        )  # batch
        return -torch.mean(log_likelihood)  # 1

    def score(self, emissions, tags, mask):
        seq_len, batch_size = tags.shape
        batch_range = torch.arange(batch_size)
        score = emissions[0, batch_range, tags[0]]  # batch
        for index in range(1, seq_len):
            score += (
                self.transitions[tags[index - 1], tags[index]]
                + emissions[index, batch_range, tags[index]]
            ) * mask[index]
        return score

    def normalization(self, emissions, mask):
        seq_len, batch_size, tags_num = emissions.shape
        score = emissions[0]
        for index in range(1, seq_len):
            score_ = score.view(batch_size, tags_num, 1)
            emissions_ = emissions[index].view(batch_size, 1, tags_num)
            score_ = score_ + self.transitions + emissions_  # batch x tags x tags
            score_ = torch.logsumexp(score_, dim=-2)  # batch x tags
            mask_ = mask[index].view(batch_size, 1)
            score = torch.where(mask_, score_, score)
        return torch.logsumexp(score, dim=-1)  # batch

    def decode(self, emissions, mask=None):
        batch_size, seq_len, tags_num = emissions.shape
        if mask is None:
            mask = torch.ones(
                (batch_size, seq_len),
                dtype=torch.bool,
                device=emissions.device
            )

        emissions = emissions.transpose(1, 0)
        mask = mask.transpose(1, 0)

        history = []
        score = emissions[0]
        for index in range(1, seq_len):
            score_ = score.view(batch_size, tags_num, 1)
            emissions_ = emissions[index].view(batch_size, 1, tags_num)
            score_ = score_ + self.transitions + emissions_  # batch x tags x tags
            score_, indexes = torch.max(score_, dim=-2)  # batch x tags
            mask_ = mask[index].view(batch_size, 1)
            score = torch.where(mask_, score_, score)
            history.append(indexes)

        sizes = mask.sum(0) - 1
        batch = []
        for index in range(batch_size):
            best = score[index].argmax()
            tags = [best]
            size = sizes[index]
            for indexes in reversed(history[:size]):
                best = indexes[index][best]
                tags.append(best)
            tags.reverse()
            batch.append(torch.tensor(tags))
        return batch





from os.path import exists, abspath

from .io import load_json

TRAIN = 'train'
DEV = 'dev'
TEST = 'test'

#########
#  DEVICE
#########

CUDA0 = 'cuda:0'
CUDA1 = 'cuda:1'
CUDA2 = 'cuda:2'
CUDA3 = 'cuda:3'
CPU = 'cpu'

#########
#   VOCAB
#########

UNK = '<unk>'
PAD = '<pad>'
CLS = '<cls>'
SEP = '<sep>'
MASK = '<mask>'

WORD = 'word'
SHAPE = 'shape'
TAG = 'tag'
REL = 'rel'

#########
#  BIO
#########

B = 'B'
I = 'I'  # noqa E741
O = 'O'  # noqa

PER = 'PER'
LOC = 'LOC'
ORG = 'ORG'

#########
#  CONFIG
#########

config = {}
path = abspath('../slovnet.json')

if exists(path):
    config = load_json(path)

#########
#  S3
#########

S3_KEY_ID = config.get('s3_key_id')
S3_KEY = config.get('s3_key')
S3_BUCKET = 'natasha-slovnet'
S3_REGION = 'us-east-1'
S3_ENDPOINT = 'https://storage.yandexcloud.net'




from slovnet.markup import SyntaxMarkup
from slovnet.mask import split_masked

from .base import Infer


class SyntaxDecoder:
    def __init__(self, rels_vocab):
        self.rels_vocab = rels_vocab

    def __call__(self, preds):
        for pred in preds:
            head_ids, rel_ids = pred
            ids = [str(_ + 1) for _ in range(len(head_ids))]
            head_ids = [str(_) for _ in head_ids.tolist()]
            rels = [self.rels_vocab.decode(_) for _ in rel_ids]
            yield ids, head_ids, rels


class SyntaxInfer(Infer):
    def process(self, inputs):
        for input in inputs:
            input = input.to(self.model.device)

            pred = self.model(input.word_id, input.shape_id, input.pad_mask)
            mask = ~input.pad_mask

            head_id = self.model.head.decode(pred.head_id, mask)
            head_id = split_masked(head_id, mask)

            rel_id = self.model.rel.decode(pred.rel_id, mask)
            rel_id = split_masked(rel_id, mask)

            yield from zip(head_id, rel_id)

    def __call__(self, items):
        inputs = self.encoder(items)
        preds = self.process(inputs)
        preds = self.decoder(preds)

        for item, pred in zip(items, preds):
            ids, head_ids, rels = pred
            tuples = zip(ids, item, head_ids, rels)
            yield SyntaxMarkup.from_tuples(tuples)




from os import getenv

import logging
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)-15s %(message)s'
)
log = logging.info

from aiohttp import web

import torch
torch.set_grad_enabled(False)

from slovnet.const import PER, LOC, ORG, CUDA0
from slovnet.vocab import (
    BERTVocab,
    BIOTagsVocab
)
from slovnet.model.bert import (
    RuBERTConfig,
    BERTEmbedding,
    BERTEncoder,
    BERTNERHead,
    BERTNER
)
from slovnet.encoders.bert import BERTInferEncoder
from slovnet.infer.bert import BERTNERInfer, BERTTagDecoder


WORDS_VOCAB = getenv('WORDS_VOCAB', 'vocab.txt')
EMB = getenv('EMB', 'emb.pt')
ENCODER = getenv('ENCODER', 'encoder.pt')
NER = getenv('NER', 'ner.pt')

DEVICE = getenv('DEVICE', CUDA0)
SEQ_LEN = int(getenv('SEQ_LEN', 256))
BATCH_SIZE = int(getenv('BATCH_SIZE', 64))

HOST = getenv('HOST', '0.0.0.0')
PORT = int(getenv('PORT', 8080))
MB = 1024 * 1024
MAX_SIZE = int(getenv('MAX_SIZE', 100 * MB))


log('Load words vocab: %r' % WORDS_VOCAB)
words_vocab = BERTVocab.load(WORDS_VOCAB)
tags_vocab = BIOTagsVocab([PER, LOC, ORG])

config = RuBERTConfig()
emb = BERTEmbedding.from_config(config)
encoder = BERTEncoder.from_config(config)
ner = BERTNERHead(config.emb_dim, len(tags_vocab))
model = BERTNER(emb, encoder, ner)
model.eval()

log('Load emb: %r' % EMB)
model.emb.load(EMB)
log('Load encoder: %r' % ENCODER)
model.encoder.load(ENCODER)
log('Load ner: %r' % NER)
model.head.load(NER)
log('Device: %r' % DEVICE)
model = model.to(DEVICE)

log('Seq len: %r' % SEQ_LEN)
log('Batch size: %r' % BATCH_SIZE)
encoder = BERTInferEncoder(
    words_vocab,
    seq_len=SEQ_LEN, batch_size=BATCH_SIZE
)
decoder = BERTTagDecoder(tags_vocab)
infer = BERTNERInfer(model, encoder, decoder)


async def handle(request):
    chunk = await request.json()
    log('Post chunk size: %r' % len(chunk))
    markups = list(infer(chunk))

    spans = sum(len(_.spans) for _ in markups)
    log('Infer spans: %r', spans)

    data = [_.as_json for _ in markups]
    return web.json_response(data)


log('Max size: %r' % (MAX_SIZE // MB))
app = web.Application(client_max_size=MAX_SIZE)
app.add_routes([web.post('/', handle)])

web.run_app(app, host=HOST, port=PORT)




import torch

from slovnet.record import Record
from slovnet.pad import pad_sequence
from slovnet.chop import chop, chop_drop
from slovnet.batch import Batch
from slovnet.mask import Masked, pad_masked
from slovnet.bert import bert_subs

from .buffer import ShuffleBuffer, SortBuffer


##########
#
#   MLM
#
########


class BERTMLMTrainEncoder:
    def __init__(self, vocab,
                 seq_len=512, batch_size=8, shuffle_size=1,
                 mask_prob=0.15):
        self.vocab = vocab
        self.seq_len = seq_len
        self.batch_size = batch_size

        self.shuffle = ShuffleBuffer(shuffle_size)

        self.mask_prob = mask_prob

    def items(self, texts):
        for text in texts:
            subs = bert_subs(text, self.vocab)
            for sub in subs:
                yield self.vocab.encode(sub)

    def seqs(self, items):
        for chunk in chop_drop(items, self.seq_len - 2):
            yield [self.vocab.cls_id] + chunk + [self.vocab.sep_id]

    def mask(self, input):
        prob = torch.full(input.shape, self.mask_prob)

        spec = (input == self.vocab.cls_id) | (input == self.vocab.sep_id)
        prob.masked_fill_(spec, 0)  # do not mask cls, sep

        return torch.bernoulli(prob).bool()

    def batch(self, chunk):
        input = torch.tensor(chunk).long()
        target = input.clone()

        mask = self.mask(input)
        input[mask] = self.vocab.mask_id

        return Batch(input, Masked(target, mask))

    def __call__(self, texts):
        items = self.items(texts)
        seqs = self.seqs(items)
        seqs = self.shuffle(seqs)
        chunks = chop(seqs, self.batch_size)
        for chunk in chunks:
            yield self.batch(chunk)


#########
#
#   NER
#
######


class BERTNERTrainEncoder:
    def __init__(self, words_vocab, tags_vocab,
                 seq_len=512, batch_size=8, shuffle_size=1):
        self.words_vocab = words_vocab
        self.tags_vocab = tags_vocab

        self.seq_len = seq_len
        self.batch_size = batch_size

        self.shuffle = ShuffleBuffer(shuffle_size)

    def items(self, markups):
        for markup in markups:
            for token in markup.tokens:
                subs = bert_subs(token.text, self.words_vocab)
                tag_id = self.tags_vocab.encode(token.tag)
                for index, sub in enumerate(subs):
                    yield (
                        self.words_vocab.encode(sub),
                        tag_id,
                        index == 0  # use first subtoken
                    )

    def seqs(self, items):
        cls = (self.words_vocab.cls_id, self.tags_vocab.pad_id, False)
        sep = (self.words_vocab.sep_id, self.tags_vocab.pad_id, False)
        for chunk in chop_drop(items, self.seq_len - 2):
            yield [cls] + chunk + [sep]

    def batch(self, chunk):
        chunk = torch.tensor(chunk)  # batch x seq x (word, mask, tag)
        word_id, tag_id, mask = chunk.unbind(-1)
        word_id, tag_id, mask = word_id.long(), tag_id.long(), mask.bool()

        input = Masked(word_id, mask)
        target = Masked(
            pad_masked(tag_id, input.mask),
            pad_masked(input.mask, input.mask)
        )

        return Batch(input, target)

    def __call__(self, markups):
        items = self.items(markups)
        seqs = self.seqs(items)
        seqs = self.shuffle(seqs)
        chunks = chop(seqs, self.batch_size)
        for chunk in chunks:
            yield self.batch(chunk)


###########
#
#   MORPH
#
#######


class BERTMorphTrainEncoder:
    def __init__(self, words_vocab, tags_vocab,
                 seq_len=512, batch_size=8, shuffle_size=1):
        self.words_vocab = words_vocab
        self.tags_vocab = tags_vocab

        self.seq_len = seq_len
        self.batch_size = batch_size

        self.shuffle = ShuffleBuffer(shuffle_size)

    def items(self, markups):
        for markup in markups:
            for token in markup.tokens:
                subs = bert_subs(token.text, self.words_vocab)
                tag_id = self.tags_vocab.encode(token.tag)
                for index, sub in enumerate(subs):
                    yield (
                        self.words_vocab.encode(sub),
                        tag_id,
                        index == 0
                    )

    def seqs(self, items):
        cls = (self.words_vocab.cls_id, self.tags_vocab.pad_id, False)
        sep = (self.words_vocab.sep_id, self.tags_vocab.pad_id, False)
        for chunk in chop_drop(items, self.seq_len - 2):
            yield [cls] + chunk + [sep]

    def batch(self, chunk):
        chunk = torch.tensor(chunk)  # batch x seq x (word, mask, tag)
        word_id, tag_id, mask = chunk.unbind(-1)
        word_id, tag_id, mask = word_id.long(), tag_id.long(), mask.bool()

        input = Masked(word_id, mask)
        target = Masked(
            pad_masked(tag_id, input.mask),
            pad_masked(input.mask, input.mask)
        )

        return Batch(input, target)

    def __call__(self, markups):
        items = self.items(markups)
        seqs = self.seqs(items)
        seqs = self.shuffle(seqs)
        chunks = chop(seqs, self.batch_size)
        for chunk in chunks:
            yield self.batch(chunk)


########
#
#   SYNTAX
#
####


ROOT_ID = '0'


class BERTSyntaxTrainItem(Record):
    __attributes__ = ['word_ids', 'mask', 'head_ids', 'rel_ids']

    def __len__(self):
        return len(self.rel_ids)


class BERTSyntaxInput(Record):
    __attributes__ = ['word_id', 'word_mask', 'pad_mask']


class BERTSyntaxTarget(Record):
    __attributes__ = ['head_id', 'rel_id', 'mask']


class BERTSyntaxTrainEncoder:
    def __init__(self, words_vocab, rels_vocab,
                 seq_len=512, batch_size=8,
                 sort_size=1):
        self.words_vocab = words_vocab
        self.rels_vocab = rels_vocab

        self.seq_len = seq_len
        self.batch_size = batch_size

        self.sort = SortBuffer(sort_size, key=lambda _: len(_.tokens))

    def item(self, markup):
        word_ids, mask, head_ids, rel_ids = [], [], [], []
        ids = {ROOT_ID: 0}

        for index, token in enumerate(markup.tokens, 1):
            ids[token.id] = index
            head_ids.append(token.head_id)

            rel_id = self.rels_vocab.encode(token.rel)
            rel_ids.append(rel_id)

            subs = bert_subs(token.text, self.words_vocab)
            for index, sub in enumerate(subs):
                word_id = self.words_vocab.encode(sub)
                word_ids.append(word_id)
                mask.append(index == 0)

        word_ids = [self.words_vocab.cls_id] + word_ids + [self.words_vocab.sep_id]
        mask = [False] + mask + [False]
        head_ids = [ids[_] for _ in head_ids]
        return BERTSyntaxTrainItem(word_ids, mask, head_ids, rel_ids)

    def batch(self, items):
        word_id, mask, head_id, rel_id = [], [], [], []
        for item in items:
            word_id.append(torch.tensor(item.word_ids, dtype=torch.long))
            mask.append(torch.tensor(item.mask, dtype=torch.bool))
            head_id.append(torch.tensor(item.head_ids, dtype=torch.long))
            rel_id.append(torch.tensor(item.rel_ids, dtype=torch.long))

        word_id = pad_sequence(word_id, fill=self.words_vocab.pad_id)
        word_mask = pad_sequence(mask, fill=False)
        pad_mask = word_id == self.words_vocab.pad_id
        input = BERTSyntaxInput(word_id, word_mask, pad_mask)

        head_id = pad_sequence(head_id)
        rel_id = pad_sequence(rel_id, fill=self.rels_vocab.pad_id)
        mask = rel_id != self.rels_vocab.pad_id
        target = BERTSyntaxTarget(head_id, rel_id, mask)

        return Batch(input, target)

    def __call__(self, markups):
        markups = self.sort(markups)
        items = (self.item(_) for _ in markups)
        # 0.02% sents longer then 128, just drop them
        items = (_ for _ in items if len(_.word_ids) <= self.seq_len)
        chunks = chop(items, self.batch_size)
        for chunk in chunks:
            yield self.batch(chunk)


#########
#
#   INFER
#
########


class BERTInferInput(Record):
    __attributes__ = ['word_id', 'word_mask', 'pad_mask']


class BERTInferEncoder:
    def __init__(self, words_vocab,
                 seq_len=128, batch_size=8):
        self.words_vocab = words_vocab

        self.seq_len = seq_len
        self.batch_size = batch_size

    def item(self, item):
        word_ids, mask = [], []
        for token in item.tokens:
            for index, sub in enumerate(token.subs):
                word_id = self.words_vocab.encode(sub)
                word_ids.append(word_id)
                mask.append(index == 0)

        size = self.seq_len - 2
        return (
            [self.words_vocab.cls_id] + word_ids[:size] + [self.words_vocab.sep_id],
            [False] + mask[:size] + [False]
        )

    def input(self, items):
        word_id, word_mask = [], []
        for word_ids, mask in items:
            word_id.append(torch.tensor(word_ids, dtype=torch.long))
            word_mask.append(torch.tensor(mask, dtype=torch.bool))
        word_id = pad_sequence(word_id, self.words_vocab.pad_id)
        word_mask = pad_sequence(word_mask, False)
        pad_mask = word_id == self.words_vocab.pad_id
        return BERTInferInput(word_id, word_mask, pad_mask)

    def __call__(self, items):
        items = (self.item(_) for _ in items)
        chunks = chop(items, self.batch_size)
        for chunk in chunks:
            yield self.input(chunk)




import numpy as np

from slovnet.record import Record, parse_annotation
from slovnet.visitor import Visitor

from .mask import fill_masked


class Weight(Record):
    __attributes__ = ['shape', 'dtype', 'array']

    def empty(self):
        return self.replace(array=None)

    @property
    def is_empty(self):
        return self.array is None

    @property
    def is_id(self):
        return type(self.array) is int


class Module(Record):
    def separate_arrays(self):
        visitor = SeparateArraysVisitor()
        scheme = visitor(self)
        return visitor.arrays, scheme

    def inject_arrays(self, arrays):
        visitor = InjectArraysVisitor(arrays)
        return visitor(self)

    def strip_navec(self):
        visitor = StripNavecVisitor()
        return visitor(self)

    def inject_navec(self, navec):
        visitor = InjectNavecVisitor(navec)
        return visitor(self)

    @property
    def weights(self):
        visitor = WeightsVisitor()
        visitor(self)
        return visitor.weights


class Linear(Module):
    __attributes__ = ['weight', 'bias']
    __annotations__ = {
        'weight': Weight,
        'bias': Weight
    }

    def __init__(self, weight, bias):
        self.weight = weight
        self.bias = bias
        self.in_dim, self.out_dim = self.weight.shape

    def __call__(self, input):
        shape = input.shape
        input = input.reshape(-1, self.in_dim)
        output = np.matmul(input, self.weight.array) + self.bias.array

        shape = shape[:-1] + (self.out_dim,)
        return output.reshape(*shape)


class Conv1d(Module):
    __attributes__ = [
        'weight',  # filters x in x kernel
        'bias',  # filters
        'padding'
    ]
    __annotations__ = {
        'weight': Weight,
        'bias': Weight
    }

    def __call__(self, input):
        input = np.pad(
            input,
            # batch no pad, in no pad, pad seq
            ((0, 0), (0, 0), (self.padding, self.padding)),
            mode='constant', constant_values=0
        )
        input = np.ascontiguousarray(input)

        batch_size, in_dim, seq_len = input.shape
        batch_stride, in_stride, seq_stride = input.strides
        unit_stride = input.data.itemsize
        filters_count, in_dim, kernel_size = self.weight.shape
        windows_count = seq_len - kernel_size + 1

        # populate conv windows
        windows = np.ndarray(
            (batch_size, windows_count, in_dim, kernel_size),
            dtype=input.dtype,
            buffer=input.data,
            strides=(batch_stride, unit_stride, in_stride, seq_stride)
        )

        # conv
        windows = windows.reshape(batch_size * windows_count, in_dim * kernel_size)
        weight = self.weight.array.reshape(filters_count, in_dim * kernel_size)
        output = np.matmul(windows, weight.T) + self.bias.array
        output = output.reshape(batch_size, windows_count, filters_count)

        # as in torch
        return output.swapaxes(2, 1)  # batch x filters x windows


class ReLU(Module):
    def __call__(self, input):
        return input.clip(0)


class BatchNorm1d(Module):
    __attributes__ = ['weight', 'bias', 'mean', 'std']
    __annotations__ = {
        'weight': Weight,
        'bias': Weight,
        'mean': Weight,
        'std': Weight
    }

    def __call__(self, input):
        # input is N x C x L, do ops on C
        input = input.swapaxes(2, 1)
        output = (
            (input - self.mean.array)
            / self.std.array
            * self.weight.array
            + self.bias.array
        )
        return output.swapaxes(2, 1)  # recover shape


#######
#
#   CRF
#
#######


class CRF(Module):
    __attributes__ = ['transitions']
    __annotations__ = {
        'transitions': Weight
    }

    def decode(self, emissions, mask):
        batch_size, seq_len, tags_num = emissions.shape
        emissions = emissions.swapaxes(1, 0)
        mask = mask.swapaxes(1, 0)

        history = []
        score = emissions[0]  # batch x tags
        for index in range(1, seq_len):
            score_ = score.reshape(batch_size, tags_num, 1)
            emissions_ = emissions[index].reshape(batch_size, 1, tags_num)
            score_ = score_ + self.transitions.array + emissions_  # batch x tags x tags
            indexes = score_.argmax(-2)
            # https://stackoverflow.com/questions/20128837/get-indices-of-numpy-argmax-elements-over-an-axis
            batch_indexes, tags_indexes = np.indices(indexes.shape)
            score_ = score_[batch_indexes, indexes, tags_indexes]

            mask_ = mask[index].reshape(batch_size, 1)
            score = np.where(mask_, score_, score)
            history.append(indexes)

        sizes = mask.sum(0) - 1
        batch = []
        for index in range(batch_size):
            best = score[index].argmax()
            tags = [best]
            size = sizes[index]
            for indexes in reversed(history[:size]):
                best = indexes[index][best]
                tags.append(best)
            tags.reverse()
            batch.append(np.array(tags))

        return batch


########
#
#   EMB
#
#######


class Embedding(Module):
    __attributes__ = ['weight']
    __annotations__ = {
        'weight': Weight
    }

    def __init__(self, weight):
        self.weight = weight
        _, self.dim = self.weight.shape

    def __call__(self, input):
        shape = input.shape
        input = input.flatten()
        weight = self.weight.array[input]
        return weight.reshape(*shape, self.dim)


class NavecEmbedding(Embedding):
    __attributes__ = ['id', 'indexes', 'codes']
    __annotations__ = {
        'indexes': Weight,
        'codes': Weight
    }

    def __init__(self, id, indexes, codes):
        self.id = id
        self.indexes = indexes
        self.codes = codes

        qdim, centroids, chunk = codes.shape
        self.dim = qdim * chunk
        self.qdims = np.arange(qdim)

    def __call__(self, input):
        shape = input.shape
        input = input.flatten()
        indexes = self.indexes.array[input]
        output = self.codes.array[self.qdims, indexes]
        return output.reshape(*shape, self.dim)


class WordShapeEmbedding(Module):
    __attributes__ = ['word', 'shape']
    __annotations__ = {
        'word': NavecEmbedding,
        'shape': Embedding
    }

    def __call__(self, word_id, shape_id):
        word = self.word(word_id)
        shape = self.shape(shape_id)
        return np.concatenate((word, shape), axis=-1)


######
#
#   CNN
#
#######


class CNNEncoderLayer(Module):
    __attributes__ = ['conv', 'relu', 'norm']
    __annotations__ = {
        'conv': Conv1d,
        'relu': ReLU,
        'norm': BatchNorm1d
    }

    def __call__(self, input):
        x = self.conv(input)
        x = self.relu(x)
        return self.norm(x)


class CNNEncoder(Module):
    __attributes__ = ['layers']
    __annotations__ = {
        'layers': [CNNEncoderLayer]
    }

    def __call__(self, input, mask):
        input = np.swapaxes(input, 2, 1)
        mask = np.expand_dims(mask, axis=1)

        for layer in self.layers:
            input = layer(input)
            size = input.shape[1]
            input[mask.repeat(size, axis=1)] = 0

        return np.swapaxes(input, 2, 1)


#######
#
#   TAG
#
######


class NERHead(Module):
    __attributes__ = ['proj', 'crf']
    __annotations__ = {
        'proj': Linear,
        'crf': CRF
    }

    def __call__(self, input):
        return self.proj(input)


class MorphHead(Module):
    __attributes__ = ['proj']
    __annotations__ = {
        'proj': Linear
    }

    def __call__(self, input):
        return self.proj(input)

    def decode(self, pred):
        return pred.argmax(-1)


class Tag(Module):
    __attributes__ = ['emb', 'encoder', 'head']

    def __call__(self, word_id, shape_id, pad_mask):
        x = self.emb(word_id, shape_id)
        x = self.encoder(x, pad_mask)
        return self.head(x)


class NER(Tag):
    __annotations__ = {
        'emb': WordShapeEmbedding,
        'encoder': CNNEncoder,
        'head': NERHead
    }


class Morph(Tag):
    __annotations__ = {
        'emb': WordShapeEmbedding,
        'encoder': CNNEncoder,
        'head': MorphHead
    }


########
#
#  SYNTAX
#
######


class FF(Module):
    __attributes__ = ['proj', 'relu']
    __annotations__ = {
        'proj': Linear,
        'relu': ReLU
    }

    def __call__(self, input):
        x = self.proj(input)
        return self.relu(x)


def append_root(input, root):
    batch_size, _, emb_dim = input.shape
    root = np.tile(root, batch_size)
    root = root.reshape(batch_size, 1, emb_dim)
    return np.concatenate((root, input), axis=1)


def strip_root(input):
    return input[:, 1:, :]


def append_root_mask(mask):
    return np.pad(
        mask,
        [(0, 0), (1, 0)],  # no pad for batch, pad left seq
        mode='constant', constant_values=True
    )


def matmul_mask(mask):
    mask = np.expand_dims(mask, axis=-2)
    return np.matmul(mask.swapaxes(-2, -1), mask)


class SyntaxHead(Module):
    __attributes__ = ['head', 'tail', 'root', 'kernel']
    __annotations__ = {
        'head': FF,
        'tail': FF,
        'root': Weight,
        'kernel': Weight
    }

    def decode(self, pred, mask):
        mask = append_root_mask(mask)
        mask = matmul_mask(mask)
        mask = strip_root(mask)

        pred = fill_masked(pred, ~mask, pred.min())
        return pred.argmax(-1)

    def __call__(self, input):
        input = append_root(input, self.root.array)
        head = self.head(input)
        tail = self.tail(input)

        x = np.matmul(head, self.kernel.array)
        x = np.matmul(x, tail.swapaxes(-2, -1))
        return strip_root(x)


def gather_head(input, root, index):
    batch_size, seq_len, emb_dim = input.shape
    input = append_root(input, root)

    zero = np.zeros((batch_size, 1), dtype=np.int_)
    index = np.concatenate((zero, index), axis=-1)
    # flatten input, absolute indexing
    input = input.reshape(-1, emb_dim)  # batch * seq x dim
    offset = np.arange(batch_size) * (seq_len + 1)
    index = offset[np.newaxis].T + index
    input = input[index]

    return strip_root(input)


class SyntaxRel(Module):
    __attributes__ = ['head', 'tail', 'root', 'kernel']
    __annotations__ = {
        'head': FF,
        'tail': FF,
        'root': Weight,
        'kernel': Weight
    }

    def decode(self, pred, mask):
        _, _, rel_dim = pred.shape
        mask = np.expand_dims(mask, axis=-1)
        mask = np.repeat(mask, rel_dim, axis=-1)

        pred = fill_masked(pred, ~mask, pred.min())
        return pred.argmax(-1)

    def __call__(self, input, head_id):
        head = self.head(gather_head(input, self.root.array, head_id))
        tail = self.tail(input)

        batch_size, seq_len, _ = input.shape
        hidden_dim, dim = self.kernel.shape
        rel_dim = dim // hidden_dim

        x = np.matmul(head, self.kernel.array)  # batch x seq x hidden * rel
        x = x.reshape(batch_size, seq_len, rel_dim, hidden_dim)
        x = np.matmul(x, tail.reshape(batch_size, seq_len, hidden_dim, 1))
        return x.reshape(batch_size, seq_len, rel_dim)


class SyntaxPred(Record):
    __attributes__ = ['head_id', 'rel_id']


class Syntax(Module):
    __attributes__ = ['emb', 'encoder', 'head', 'rel']
    __annotations__ = {
        'emb': WordShapeEmbedding,
        'encoder': CNNEncoder,
        'head': SyntaxHead,
        'rel': SyntaxRel
    }

    def __call__(self, word_id, shape_id, pad_mask):
        x = self.emb(word_id, shape_id)
        x = self.encoder(x, pad_mask)

        head_id = self.head(x)
        target_head_id = self.head.decode(head_id, ~pad_mask)
        rel_id = self.rel(x, target_head_id)
        return SyntaxPred(head_id, rel_id)


#######
#
#   VISITOR
#
######


class ModuleVisitor(Visitor):
    def visit_Weight(self, item):
        return item

    def visit_Module(self, item):
        args = []
        for key in item.__attributes__:
            value = getattr(item, key)
            annotation = item.__annotations__.get(key)
            if annotation and value is not None:
                _, repeatable, _ = parse_annotation(annotation)
                if repeatable:
                    value = [self.visit(_) for _ in value]
                else:
                    value = self.visit(value)
            args.append(value)
        return type(item)(*args)


class SeparateArraysVisitor(ModuleVisitor):
    def __init__(self):
        self.arrays = {}

    def visit_Weight(self, item):
        if item.is_empty:
            return item

        id = len(self.arrays)
        self.arrays[id] = item.array
        return item.replace(array=id)


class InjectArraysVisitor(ModuleVisitor):
    def __init__(self, arrays):
        self.arrays = arrays

    def visit_Weight(self, item):
        if not item.is_id:
            return item

        return item.replace(
            array=self.arrays[item.array]
        )


class StripNavecVisitor(ModuleVisitor):
    def visit_NavecEmbedding(self, item):
        return item.replace(
            indexes=item.indexes.empty(),
            codes=item.codes.empty()
        )


class InjectNavecVisitor(ModuleVisitor):
    def __init__(self, navec):
        self.navec = navec

    def visit_NavecEmbedding(self, item):
        id = self.navec.meta.id
        if item.id != id:
            raise ValueError('Expected id=%r, got %r' % (item.id, id))

        pq = self.navec.pq
        return item.replace(
            indexes=item.indexes.replace(array=pq.indexes),
            codes=item.codes.replace(array=pq.codes)
        )


class WeightsVisitor(ModuleVisitor):
    def __init__(self):
        self.weights = []

    def visit_Weight(self, item):
        self.weights.append(item)
        return item




import re
from os import listdir, makedirs
from os.path import exists, join, expanduser
from random import seed, shuffle

from tqdm.notebook import tqdm as log_progress

import torch
from torch import optim

from apex import amp
O2 = 'O2'

from corus import (
    load_buriy_news,
    load_taiga_fontanka,
    load_ods_gazeta,
    load_ods_interfax,
    load_lenta
)

from slovnet.io import (
    load_lines,
    dump_lines
)
from slovnet.s3 import S3
from slovnet.board import TensorBoard
from slovnet.const import CUDA0

from slovnet.model.bert import (
    RuBERTConfig,
    BERTEmbedding,
    BERTEncoder,
    BERTMLMHead,
    BERTMLM
)
from slovnet.vocab import BERTVocab
from slovnet.encoders.bert import BERTMLMTrainEncoder
from slovnet.score import (
    MLMScoreMeter,
    score_mlm_batch,
    score_mlm_batches
)
from slovnet.loss import masked_flatten_cross_entropy


DATA_DIR = 'data'
MODEL_DIR = 'model'
RUBERT_DIR = 'rubert'
RAW_DIR = join(DATA_DIR, 'raw')

TRAIN = join(DATA_DIR, 'train.txt')
TEST = join(DATA_DIR, 'test.txt')

S3_DIR = '01_bert_news'
S3_TRAIN = join(S3_DIR, TRAIN)
S3_TEST = join(S3_DIR, TEST)

VOCAB = 'vocab.txt'
EMB = 'emb.pt'
ENCODER = 'encoder.pt'
MLM = 'mlm.pt'

RUBERT_VOCAB = join(RUBERT_DIR, VOCAB)
RUBERT_EMB = join(RUBERT_DIR, EMB)
RUBERT_ENCODER = join(RUBERT_DIR, ENCODER)
RUBERT_MLM = join(RUBERT_DIR, MLM)

S3_RUBERT_VOCAB = join(S3_DIR, RUBERT_VOCAB)
S3_RUBERT_EMB = join(S3_DIR, RUBERT_EMB)
S3_RUBERT_ENCODER = join(S3_DIR, RUBERT_ENCODER)
S3_RUBERT_MLM = join(S3_DIR, RUBERT_MLM)

MODEL_EMB = join(MODEL_DIR, EMB)
MODEL_ENCODER = join(MODEL_DIR, ENCODER)
MODEL_MLM = join(MODEL_DIR, MLM)

S3_MODEL_EMB = join(S3_DIR, MODEL_EMB)
S3_MODEL_ENCODER = join(S3_DIR, MODEL_ENCODER)
S3_MODEL_MLM = join(S3_DIR, MODEL_MLM)

BOARD_NAME = '01_bert_news'
RUNS_DIR = 'runs'

TRAIN_BOARD = '01_train'
TEST_BOARD = '02_test'

DEVICE = CUDA0


def every(step, period):
    return step > 0 and step % period == 0


def process_batch(model, criterion, batch):
    pred = model(batch.input)
    loss = criterion(pred, batch.target.value, batch.target.mask)
    return batch.processed(loss, pred)


def infer_batches(model, criterion, batches):
    training = model.training
    model.eval()
    with torch.no_grad():
        for batch in batches:
            yield process_batch(model, criterion, batch)
    model.train(training)




from torch import nn

from .base import Module


class CNNEncoderLayer(Module):
    def __init__(self, in_dim, out_dim, kernel_size):
        super(CNNEncoderLayer, self).__init__()

        padding = (kernel_size - 1) // 2
        self.conv = nn.Conv1d(
            in_dim, out_dim, kernel_size,
            padding=padding
        )
        self.relu = nn.ReLU()
        self.norm = nn.BatchNorm1d(out_dim)

    def __call__(self, input):
        x = self.conv(input)
        x = self.relu(x)
        return self.norm(x)


def gen_cnn_encoder_layers(input_dim, layer_dims, kernel_size):
    dims = [input_dim] + layer_dims
    for index in range(1, len(dims)):
        in_dim = dims[index - 1]
        out_dim = dims[index]
        yield CNNEncoderLayer(in_dim, out_dim, kernel_size)


class CNNEncoder(Module):
    def __init__(self, input_dim, layer_dims, kernel_size):
        super(CNNEncoder, self).__init__()

        layers = gen_cnn_encoder_layers(input_dim, layer_dims, kernel_size)
        self.layers = nn.ModuleList(layers)
        self.dim = layer_dims[-1]

    def forward(self, input, mask=None):  # batch x seq x emb
        input = input.transpose(2, 1)  # batch x emb x seq

        if mask is not None:
            mask = mask.unsqueeze(1)  # batch x 1 x seq

        for layer in self.layers:
            input = layer(input)  # batch x dim x seq

            if mask is not None:
                input[mask.expand_as(input)] = 0

        return input.transpose(2, 1)  # batch x seq x dim




from os import getenv, environ
from os.path import exists, join, expanduser
from random import seed, sample, randint, uniform
from subprocess import run

from tqdm.notebook import tqdm as log_progress

import torch
from torch import optim

from naeval.ner.datasets import (
    load_factru,
    load_ne5,
)

from slovnet.s3 import S3
from slovnet.io import (
    format_jl,
    parse_jl,

    load_gz_lines,
    dump_gz_lines
)
from slovnet.board import (
    TensorBoard,
    LogBoard,
    MultiBoard
)
from slovnet.const import (
    TRAIN, TEST,
    PER, LOC, ORG,
    CUDA0,
)
from slovnet.token import tokenize

from slovnet.model.bert import (
    RuBERTConfig,
    BERTEmbedding,
    BERTEncoder,
    BERTNERHead,
    BERTNER
)
from slovnet.markup import (
    SpanMarkup,
    show_span_markup
)
from slovnet.vocab import BERTVocab, BIOTagsVocab
from slovnet.encoders.bert import BERTNERTrainEncoder, BERTInferEncoder
from slovnet.score import (
    NERBatchScore,
    NERScoreMeter,
    score_ner_batch
)
from slovnet.mask import (
    Masked,
    split_masked,
    pad_masked
)

from slovnet.infer.bert import BERTNERInfer, BERTTagDecoder

DATA_DIR = 'data'
MODEL_DIR = 'model'
BERT_DIR = 'bert'

RAW_DIR = join(DATA_DIR, 'raw')

CORUS_NE5 = join(RAW_DIR, 'Collection5')
CORUS_FACTRU = join(RAW_DIR, 'factRuEval-2016-master')

NE5 = join(DATA_DIR, 'ne5.jl.gz')
FACTRU = join(DATA_DIR, 'factru.jl.gz')

S3_DIR = '02_bert_ner'
S3_NE5 = join(S3_DIR, NE5)
S3_FACTRU = join(S3_DIR, FACTRU)

VOCAB = 'vocab.txt'
EMB = 'emb.pt'
ENCODER = 'encoder.pt'
NER = 'ner.pt'

BERT_VOCAB = join(BERT_DIR, VOCAB)
BERT_EMB = join(BERT_DIR, EMB)
BERT_ENCODER = join(BERT_DIR, ENCODER)

S3_RUBERT_DIR = '01_bert_news/rubert'
S3_MLM_DIR = '01_bert_news/model'
S3_BERT_VOCAB = join(S3_RUBERT_DIR, VOCAB)
S3_BERT_EMB = join(S3_MLM_DIR, EMB)
S3_BERT_ENCODER = join(S3_MLM_DIR, ENCODER)

MODEL_ENCODER = join(MODEL_DIR, ENCODER)
MODEL_NER = join(MODEL_DIR, NER)

S3_MODEL_ENCODER = join(S3_DIR, MODEL_ENCODER)
S3_MODEL_NER = join(S3_DIR, MODEL_NER)

BOARD_NAME = getenv('board_name', '02_bert_ner')
RUNS_DIR = 'runs'

TRAIN_BOARD = '01_train'
TEST_BOARD = '02_test'

SEED = int(getenv('seed', 72))
DEVICE = getenv('device', CUDA0)
BERT_LR = float(getenv('bert_lr', 0.000045))
LR = float(getenv('lr', 0.0075))
LR_GAMMA = float(getenv('lr_gamma', 0.45))
EPOCHS = int(getenv('epochs', 5))

SEQ_LEN = int(getenv('SEQ_LEN', 256))
BATCH_SIZE = int(getenv('BATCH_SIZE', 64))

#####################
#
#  CUSTOM TAGS TUNING
#
############### START

CUSTOM_TUNING = False # Set this flag to true in order to use your custom dataset and tags
CUSTOM_TEXTS = join(DATA_DIR, 'custom-dataset.jl.gz') # Put your own data into the data dir
TAGS = ['CUSTOM-TAG'] if CUSTOM_TUNING else [PER, LOC, ORG] # List all your custom tags 

################# END


def process_batch(model, criterion, batch):
    input, target = batch

    pred = model(input.value)
    pred = pad_masked(pred, input.mask)
    mask = pad_masked(input.mask, input.mask)

    loss = criterion(pred, target.value, target.mask)

    pred = Masked(pred, mask)
    return batch.processed(loss, pred)




from os import getenv, environ
from os.path import exists, join, expanduser
from random import seed, shuffle, sample, randint, uniform
from subprocess import run

from tqdm.notebook import tqdm as log_progress

import torch
from torch import optim

from nerus import load_nerus

from navec import Navec

from slovnet.s3 import S3
from slovnet.io import (
    format_jl,
    parse_jl,

    load_gz_lines,
    dump_gz_lines
)
from slovnet.board import (
    TensorBoard,
    LogBoard,
    MultiBoard
)
from slovnet.const import (
    TRAIN, TEST,
    CUDA0, PAD,
    WORD, SHAPE, REL
)
from slovnet.token import tokenize
from slovnet.shape import SHAPES

from slovnet.markup import (
    SyntaxToken,
    SyntaxMarkup,
    show_syntax_markup
)
from slovnet.model.emb import (
    Embedding,
    NavecEmbedding
)
from slovnet.model.syntax import (
    SyntaxEmbedding,
    SyntaxEncoder,
    SyntaxHead,
    SyntaxRel,
    Syntax
)
from slovnet.loss import masked_flatten_cross_entropy
from slovnet.vocab import Vocab
from slovnet.encoders.syntax import SyntaxTrainEncoder
from slovnet.score import (
    SyntaxBatchScore,
    SyntaxScoreMeter,
    score_syntax_batch
)

from slovnet.exec.pack import (
    Meta,
    DumpPack
)
from slovnet import api


DATA_DIR = 'data'
MODEL_DIR = 'model'
NAVEC_DIR = 'navec'
RAW_DIR = join(DATA_DIR, 'raw')
S3_DIR = '07_syntax'

RAW_NERUS = join(RAW_DIR, 'nerus_lenta.conllu.gz')
NERUS_TOTAL = 739346

NERUS = join(DATA_DIR, 'nerus.jl.gz')
S3_NERUS = join(S3_DIR, NERUS)

NAVEC_URL = 'https://storage.yandexcloud.net/natasha-navec/packs/navec_news_v1_1B_250K_300d_100q.tar'
NAVEC = join(NAVEC_DIR, 'navec_news_v1_1B_250K_300d_100q.tar')

RELS_VOCAB = join(MODEL_DIR, 'rels_vocab.txt')
MODEL_SHAPE = join(MODEL_DIR, 'shape.pt')
MODEL_ENCODER = join(MODEL_DIR, 'encoder.pt')
MODEL_HEAD = join(MODEL_DIR, 'head.pt')
MODEL_REL = join(MODEL_DIR, 'rel.pt')

S3_RELS_VOCAB = join(S3_DIR, RELS_VOCAB)
S3_MODEL_SHAPE = join(S3_DIR, MODEL_SHAPE)
S3_MODEL_ENCODER = join(S3_DIR, MODEL_ENCODER)
S3_MODEL_HEAD = join(S3_DIR, MODEL_HEAD)
S3_MODEL_REL = join(S3_DIR, MODEL_REL)

ID = 'slovnet_syntax_news_v1'
PACK = ID + '.tar'
S3_PACK = join('packs', PACK)

BOARD_NAME = getenv('board_name', '07_syntax')
RUNS_DIR = 'runs'

TRAIN_BOARD = '01_train'
TEST_BOARD = '02_test'

SEED = int(getenv('seed', 17))
DEVICE = getenv('device', CUDA0)

SHAPE_DIM = int(getenv('shape_dim', 30))
LAYERS_NUM = int(getenv('layers_num', 3))
LAYER_DIM = int(getenv('layer_dim', 64))
KERNEL_SIZE = int(getenv('kernel_size', 3))

LR = float(getenv('lr', 0.0051))
LR_GAMMA = float(getenv('lr_gamma', 0.74))
EPOCHS = int(getenv('epochs', 3))

LAYER_DIMS = [
    LAYER_DIM * 2**_
    for _ in reversed(range(LAYERS_NUM))
]


def adapt_markup(record):
    return SyntaxMarkup([
        SyntaxToken(_.id, _.text, _.head_id, _.rel)
        for _ in record.tokens
    ])


def process_batch(model, criterion, batch):
    input, target = batch

    pred = model(
        input.word_id, input.shape_id, input.pad_mask,
        target.mask, target.head_id
    )
    loss = (
        criterion(pred.head_id, target.head_id, target.mask)
        + criterion(pred.rel_id, target.rel_id, target.mask)
    )

    pred.head_id = model.head.decode(pred.head_id, target.mask)
    pred.rel_id = model.rel.decode(pred.rel_id, target.mask)

    return batch.processed(loss, pred)




import torch

from slovnet.const import CPU


def load_model(model, path):
    model.load_state_dict(torch.load(path, map_location=CPU))


def dump_model(model, path):
    torch.save(model.state_dict(), path)


class StateMixin:
    def load(self, path):
        load_model(self, path)
        return self

    def dump(self, path):
        dump_model(self, path)




from .record import Record
from .mask import mask_like
from .bio import (
    I,
    bio_io,
    parse_bio,
    select_type_tags
)


class Acc(Record):
    __attributes__ = ['correct', 'total']

    def __init__(self, correct=0, total=0):
        self.correct = correct
        self.total = total

    def add(self, other):
        self.correct += other.correct
        self.total += other.total

    @property
    def value(self):
        if not self.total:
            return 0
        return self.correct / self.total

    def reset(self):
        self.correct = 0
        self.total = 0


class Mean(Record):
    __attributes__ = ['accum', 'count']

    def __init__(self, accum=0, count=0):
        self.accum = accum
        self.count = count

    def add(self, value):
        self.accum += value
        self.count += 1

    @property
    def value(self):
        if not self.count:
            return 0
        return self.accum / self.count

    def reset(self):
        self.accum = 0
        self.count = 0


class F1(Record):
    __attributes__ = ['prec', 'recall']

    def __init__(self, prec=None, recall=None):
        if not prec:
            prec = Acc()
        self.prec = prec
        if not recall:
            recall = Acc()
        self.recall = recall

    def add(self, other):
        self.prec.add(other.prec)
        self.recall.add(other.recall)

    @property
    def value(self):
        prec = self.prec.value
        recall = self.recall.value
        if not prec + recall:
            return 0
        return 2 * prec * recall / (prec + recall)

    def reset(self):
        self.prec.reset()
        self.recall.reset()


def topk_acc(pred, target, ks=(1, 2, 4, 8), mask=None):
    k = max(ks)
    pred = pred.topk(
        k,
        dim=-1,
        largest=True,
        sorted=True
    ).indices

    if mask is None:
        mask = mask_like(target)

    target = target[mask]
    pred = pred[mask].view(-1, k)  # restore shape

    pred = pred.t()  # k x tests
    target = target.expand_as(pred)  # k x tests

    correct = (pred == target)
    total = mask.sum().item()
    for k in ks:
        count = correct[:k].sum().item()
        yield Acc(count, total)


def acc(a, b, mask=None):
    if mask is None:
        mask = mask_like(a)

    a = a[mask]
    b = b[mask]
    correct = (a == b).sum().item()
    total = len(a)
    return Acc(correct, total)


class BatchScore(Record):
    __attributes__ = ['loss']


class ScoreMeter(Record):
    __attributes__ = ['loss']

    def extend(self, scores):
        for score in scores:
            self.add(score)


###########
#
#   MLM
#
#######


class MLMBatchScore(BatchScore):
    __attributes__ = ['loss', 'ks']

    def __init__(self, loss, ks):
        self.loss = loss
        self.ks = ks


class MLMScoreMeter(ScoreMeter):
    __attributes__ = ['loss', 'ks']

    def __init__(self, loss=None, ks=None):
        if not loss:
            loss = Mean()
        if not ks:
            ks = {}
        self.loss = loss
        self.ks = ks

    def add(self, score):
        self.loss.add(score.loss)
        for k, score in score.ks.items():
            if k not in self.ks:
                self.ks[k] = score
            else:
                self.ks[k].add(score)

    def reset(self):
        self.loss.reset()
        for k in self.ks:
            self.ks[k].reset()

    def write(self, board):
        board.add_scalar('01_loss', self.loss.value)
        for index, k in enumerate(self.ks, 2):
            key = '%02d_top%d' % (index, k)
            score = self.ks.get(k)
            if score:
                board.add_scalar(key, score.value)


def score_mlm_batch(batch, ks=(1, 2, 4, 8)):
    scores = ()
    if ks:
        scores = topk_acc(batch.pred, batch.target, ks)
    return MLMBatchScore(
        batch.loss.item(),
        ks=dict(zip(ks, scores))
    )


def score_mlm_batches(batches):
    for batch in batches:
        yield score_mlm_batch(batch)


#######
#
#   NER
#
######


class NERBatchScore(BatchScore):
    __attributes__ = ['loss', 'types']

    def __init__(self, loss, types=None):
        self.loss = loss
        if not types:
            types = {}
        self.types = types


class NERScoreMeter(ScoreMeter):
    __attributes__ = ['loss', 'types']

    def __init__(self, loss=None, types=None):
        if not loss:
            loss = Mean()
        if not types:
            types = {}
        self.loss = loss
        self.types = types

    def add(self, score):
        self.loss.add(score.loss)
        for type, score in score.types.items():
            if type not in self.types:
                self.types[type] = score
            else:
                self.types[type].add(score)

    def reset(self):
        self.loss.reset()
        for type in self.types:
            self.types[type].reset()

    def write(self, board):
        board.add_scalar('01_loss', self.loss.value)
        for index, type in enumerate(self.types, 2):
            key = '%02d_%s' % (index, type)
            score = self.types.get(type)
            if score:
                board.add_scalar(key, score.value)


def tag_f1(preds, targets, type):
    score = F1()
    preds = list(bio_io(select_type_tags(preds, type)))
    targets = list(bio_io(select_type_tags(targets, type)))
    for pred, target in zip(preds, targets):
        pred, _ = parse_bio(pred)
        target, _ = parse_bio(target)
        if pred == I:
            score.prec.total += 1
            if target == I:
                score.prec.correct += 1
        if target == I:
            score.recall.total += 1
            if pred == I:
                score.recall.correct += 1
    return score


def decode_tags(seqs, tags_vocab):
    return [
        tags_vocab.decode(id)
        for seq in seqs
        for id in seq
    ]


def score_ner_batch(batch, tags_vocab):
    score = NERBatchScore(batch.loss)
    preds = decode_tags(batch.pred, tags_vocab)
    targets = decode_tags(batch.target, tags_vocab)
    for type in tags_vocab.types:
        score.types[type] = tag_f1(preds, targets, type)
    return score


###########
#
#   MORPH
#
######


class MorphBatchScore(BatchScore):
    __attributes__ = ['loss', 'acc']

    def __init__(self, loss, acc):
        self.loss = loss
        self.acc = acc


class MorphScoreMeter(ScoreMeter):
    __attributes__ = ['loss', 'acc']

    def __init__(self, loss=None, acc=None):
        if not loss:
            loss = Mean()
        if not acc:
            acc = Acc()
        self.loss = loss
        self.acc = acc

    def add(self, score):
        self.loss.add(score.loss)
        self.acc.add(score.acc)

    def reset(self):
        self.loss.reset()
        self.acc.reset()

    def write(self, board):
        board.add_scalar('01_loss', self.loss.value)
        board.add_scalar('02_acc', self.acc.value)


def score_morph_batch(batch):
    return MorphBatchScore(
        batch.loss.item(),
        acc(batch.pred, batch.target)
    )


########
#
#   SYNTAX
#
#######


class SyntaxBatchScore(Record):
    __attributes__ = ['loss', 'uas', 'las']

    def __init__(self, loss, uas=None, las=None):
        self.loss = loss
        self.uas = uas
        self.las = las


class SyntaxScoreMeter(Record):
    __attributes__ = ['loss', 'uas', 'las']

    def __init__(self, loss=None, uas=None, las=None):
        if not loss:
            loss = Mean()
        if not uas:
            uas = Acc()
        if not las:
            las = Acc()
        self.loss = loss
        self.uas = uas
        self.las = las

    def add(self, score):
        self.loss.add(score.loss)
        self.uas.add(score.uas)
        self.las.add(score.las)

    def reset(self):
        self.loss.reset()
        self.uas.reset()
        self.las.reset()

    def write(self, board):
        board.add_scalar('01_loss', self.loss.value)
        board.add_scalar('02_uas', self.uas.value)
        board.add_scalar('03_las', self.las.value)


def uas(pred, target, mask=None):
    if mask is None:
        mask = mask_like(target)

    pred = pred[mask]
    target = target[mask]

    total = len(pred)
    correct = (pred == target).sum().item()
    return Acc(correct, total)


def las(head_pred, head_target, rel_pred, rel_target, mask=None):
    if mask is None:
        mask = mask_like(head_target)

    head_pred = head_pred[mask]
    head_target = head_target[mask]
    rel_pred = rel_pred[mask]
    rel_target = rel_target[mask]

    total = len(head_pred)
    match = (head_pred == head_target) & (rel_pred == rel_target)
    correct = match.sum().item()
    return Acc(correct, total)


def score_syntax_batch(batch):
    input, target, loss, pred = batch
    return SyntaxBatchScore(
        loss.item(),
        uas(pred.head_id, target.head_id, target.mask),
        las(
            pred.head_id, target.head_id,
            pred.rel_id, target.rel_id,
            target.mask
        )
    )




from os import getenv

import logging
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)-15s %(message)s'
)
log = logging.info

from aiohttp import web

import torch
torch.set_grad_enabled(False)

from navec import Navec

from slovnet.const import PAD, CUDA0
from slovnet.shape import SHAPES
from slovnet.vocab import Vocab
from slovnet.model.emb import (
    NavecEmbedding,
    Embedding
)
from slovnet.model.tag import (
    TagEmbedding,
    TagEncoder,
    MorphHead,
    Morph
)
from slovnet.encoders.tag import TagInferEncoder
from slovnet.infer.tag import MorphInfer, TagDecoder


NAVEC = getenv('NAVEC', 'navec.tar')
SHAPE = getenv('SHAPE', 'shape.pt')
ENCODER = getenv('ENCODER', 'encoder.pt')
MORPH = getenv('MORPH', 'morph.pt')
TAGS_VOCAB = getenv('TAGS_VOCAB', 'tags_vocab.txt')

SHAPE_DIM = 30
LAYER_DIMS = [256, 128, 64]
KERNEL_SIZE = 3

DEVICE = getenv('DEVICE', CUDA0)
BATCH_SIZE = int(getenv('BATCH_SIZE', 64))

HOST = getenv('HOST', '0.0.0.0')
PORT = int(getenv('PORT', 8080))
MB = 1024 * 1024
MAX_SIZE = int(getenv('MAX_SIZE', 100 * MB))


log('Load navec: %r' % NAVEC)
navec = Navec.load(NAVEC)

words_vocab = Vocab(navec.vocab.words)
shapes_vocab = Vocab([PAD] + SHAPES)
tags_vocab = Vocab.load(TAGS_VOCAB)

word = NavecEmbedding(navec)
shape = Embedding(
    vocab_size=len(shapes_vocab),
    dim=SHAPE_DIM,
    pad_id=shapes_vocab.pad_id
)
emb = TagEmbedding(word, shape)
encoder = TagEncoder(
    input_dim=emb.dim,
    layer_dims=LAYER_DIMS,
    kernel_size=KERNEL_SIZE,
)
morph = MorphHead(encoder.dim, len(tags_vocab))
model = Morph(emb, encoder, morph)
model.eval()

log('Load shape: %r' % SHAPE)
model.emb.shape.load(SHAPE)
log('Load encoder: %r' % ENCODER)
model.encoder.load(ENCODER)
log('Load morph: %r' % MORPH)
model.head.load(MORPH)
log('Device: %r' % DEVICE)
model = model.to(DEVICE)

log('Batch size: %r' % BATCH_SIZE)
encoder = TagInferEncoder(
    words_vocab, shapes_vocab,
    batch_size=BATCH_SIZE
)
decoder = TagDecoder(tags_vocab)
infer = MorphInfer(model, encoder, decoder)


async def handle(request):
    chunk = await request.json()
    log('Post chunk size: %r' % len(chunk))
    markups = list(infer(chunk))

    tags = sum(len(_.tags) for _ in markups)
    log('Infer tags: %r', tags)

    data = [_.as_json for _ in markups]
    return web.json_response(data)


log('Max size: %r' % (MAX_SIZE // MB))
app = web.Application(client_max_size=MAX_SIZE)
app.add_routes([web.post('/', handle)])

web.run_app(app, host=HOST, port=PORT)




from .record import Record


class Span(Record):
    __attributes__ = ['start', 'stop', 'type']

    def __init__(self, start, stop, type=None):
        self.start = start
        self.stop = stop
        self.type = type

    def offset(self, delta):
        return Span(
            self.start + delta,
            self.stop + delta,
            self.type
        )


def offset_spans(spans, delta):
    for span in spans:
        yield span.offset(delta)


def envelop_span(envelope, span):
    return envelope.start <= span.start and span.stop <= envelope.stop


def envelop_spans(envelope, spans):
    for span in spans:
        if envelop_span(envelope, span):
            yield span


def select_type_spans(spans, types):
    for span in spans:
        if span.type in types:
            yield span




from setuptools import setup, find_packages


with open('README.md') as file:
    description = file.read()


setup(
    name='slovnet',
    version='0.6.0',

    description='Deep-learning based NLP modeling for Russian language',
    long_description=description,
    long_description_content_type='text/markdown',

    url='https://github.com/natasha/slovnet',
    author='Alexander Kukushkin',
    author_email='alex@alexkuk.ru',
    license='MIT',

    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    keywords='nlp, deeplearning, russian',

    packages=find_packages(
        exclude=['tests']
    ),
    install_requires=[
        'numpy',
        'razdel',
        'navec',
    ]
)




from .record import Record
from .const import WORD, SHAPE, TAG, REL
from .chop import chop

from .exec.pack import Pack
from .exec.model import (
    Morph as MorphModel,
    NER as NERModel,
    Syntax as SyntaxModel
)
from .exec.encoders import (
    TagEncoder,
    SyntaxEncoder
)
from .exec.infer import (
    TagDecoder,
    MorphInfer,
    NERInfer,

    SyntaxDecoder,
    SyntaxInfer
)


class API(Record):
    __attributes__ = ['infer', 'batch_size']

    def navec(self, navec):
        self.infer.model = self.infer.model.inject_navec(navec)
        return self

    def map(self, items):
        for chunk in chop(items, self.batch_size):
            yield from self.infer(chunk)

    def __call__(self, item):
        return next(self.map([item]))


class NER(API):
    @classmethod
    def load(cls, path, batch_size=8):
        with Pack(path) as pack:
            meta = pack.load_meta()
            meta.check_protocol()

            model = pack.load_model(NERModel)
            arrays = dict(pack.load_arrays(model.weights))

            words_vocab = pack.load_vocab(WORD)
            shapes_vocab = pack.load_vocab(SHAPE)
            tags_vocab = pack.load_vocab(TAG)

        model = model.inject_arrays(arrays)
        encoder = TagEncoder(
            words_vocab, shapes_vocab,
            batch_size
        )
        decoder = TagDecoder(tags_vocab)
        infer = NERInfer(model, encoder, decoder)

        return cls(infer, batch_size)


class Morph(API):
    @classmethod
    def load(cls, path, batch_size=8):
        with Pack(path) as pack:
            meta = pack.load_meta()
            meta.check_protocol()

            model = pack.load_model(MorphModel)
            arrays = dict(pack.load_arrays(model.weights))

            words_vocab = pack.load_vocab(WORD)
            shapes_vocab = pack.load_vocab(SHAPE)
            tags_vocab = pack.load_vocab(TAG)

        model = model.inject_arrays(arrays)
        encoder = TagEncoder(
            words_vocab, shapes_vocab,
            batch_size
        )
        decoder = TagDecoder(tags_vocab)
        infer = MorphInfer(model, encoder, decoder)

        return cls(infer, batch_size)


class Syntax(API):
    @classmethod
    def load(cls, path, batch_size=8):
        with Pack(path) as pack:
            meta = pack.load_meta()
            meta.check_protocol()

            model = pack.load_model(SyntaxModel)
            arrays = dict(pack.load_arrays(model.weights))

            words_vocab = pack.load_vocab(WORD)
            shapes_vocab = pack.load_vocab(SHAPE)
            rels_vocab = pack.load_vocab(REL)

        model = model.inject_arrays(arrays)
        encoder = SyntaxEncoder(
            words_vocab, shapes_vocab,
            batch_size
        )
        decoder = SyntaxDecoder(rels_vocab)
        infer = SyntaxInfer(model, encoder, decoder)

        return cls(infer, batch_size)




import re

from razdel import sentenize as sentenize_

from .record import Record


class Sent(Record):
    __attributes__ = ['start', 'stop', 'text']


def split_lines(text):
    for match in re.finditer(r'([^\r\n]+)', text):
        start = match.start()
        stop = match.end()
        line = match.group(1)
        yield Sent(start, stop, line)


def sentenize(text):
    for line in split_lines(text):
        for sent in sentenize_(line.text):
            if not sent.text:  # '\n\t\n' for example
                continue
            yield Sent(
                sent.start + line.start,
                sent.stop + line.start,
                sent.text
            )




from os.path import join as join_path

from torch.utils.tensorboard import SummaryWriter

from .record import Record
from .log import log


class Board(Record):
    __attributes__ = ['steps']

    def __init__(self, steps=0):
        self.steps = steps

    def section(self, name):
        return BoardSection(name, self)

    def step(self):
        self.steps += 1


class TensorBoard(Board):
    __attributes__ = ['dir', 'root', 'steps']

    def __init__(self, dir, root, steps=0, flush_secs=1):
        self.dir = dir
        self.root = root
        self.writer = SummaryWriter(
            join_path(root, dir),
            flush_secs=flush_secs
        )
        super(TensorBoard, self).__init__(steps)

    def add_scalar(self, key, value):
        self.writer.add_scalar(key, value, self.steps)


class LogBoard(Board):
    def add_scalar(self, key, value):
        log('{:>4} {:.4f} {}'.format(self.steps, value, key))


class MultiBoard(Board):
    __attributes__ = ['boards']

    def __init__(self, boards):
        self.boards = boards

    def step(self):
        for board in self.boards:
            board.step()

    def add_scalar(self, key, value):
        for board in self.boards:
            board.add_scalar(key, value)


class BoardSection(Record):
    __attributes__ = ['name', 'board']

    def prefixed(self, key):
        return '%s/%s' % (self.name, key)

    def add_scalar(self, key, value):
        self.board.add_scalar(self.prefixed(key), value)





def chop(items, size):
    buffer = []
    for item in items:
        buffer.append(item)
        if len(buffer) >= size:
            yield buffer
            buffer = []
    if buffer:
        yield buffer


def chop_drop(items, size):
    chunks = chop(items, size)
    for chunk in chunks:
        if len(chunk) < size:
            continue
        yield chunk


def chop_weighted(items, size, weight):
    buffer = []
    accum = 0
    for item in items:
        value = weight(item)
        if accum + value > size:
            yield buffer
            buffer = []
            accum = 0
        buffer.append(item)
        accum += value
    if buffer:
        yield buffer




import tarfile
from io import BytesIO

from .record import Record


class Tar(Record):
    __attributes__ = ['path']

    mode = 'r'

    def __enter__(self):
        self.tar = tarfile.open(self.path, self.mode)
        return self

    def __exit__(self, *args):
        self.tar.close()

    def open(self, name):
        member = self.tar.getmember(name)
        return self.tar.extractfile(member)

    def read(self, name):
        return self.open(name).read()

    def list(self, prefix=None):
        for member in self.tar:
            name = member.name
            if not prefix or name.startswith(prefix):
                yield name


class DumpTar(Tar):
    mode = 'w'

    def write(self, bytes, name):
        file = BytesIO(bytes)
        info = tarfile.TarInfo(name)
        info.size = len(bytes)
        self.tar.addfile(tarinfo=info, fileobj=file)





class Visitor(object):
    def resolve_method(self, item):
        for cls in item.__class__.__mro__:
            name = 'visit_' + cls.__name__
            method = getattr(self, name, None)
            if method:
                return method
        raise ValueError('no method for {type!r}'.format(
            type=type(item)
        ))

    def visit(self, item):
        return self.resolve_method(item)(item)

    def __call__(self, item):
        return self.visit(item)




from .span import Span
from .const import B, I, O


def parse_bio(tag):
    if '-' in tag:
        part, type = tag.split('-', 1)
    else:
        part = tag
        type = None
    return part, type


def format_bio(part, type):
    if not type:
        return part
    return '%s-%s' % (part, type)


##########
#
#    IO
#
#########

# assert tokens and spans are sorted
# assert spans do not overlap
# assert span bounds align with token bounds


def append_ellipsis(items, ellipsis=None):
    for item in items:
        yield item
    yield ellipsis


def spans_io(tokens, spans):
    spans = append_ellipsis(spans)
    span = next(spans)
    for token in tokens:
        part = O
        type = None
        if span:
            if token.start >= span.start:
                part = I
                type = span.type
            if token.stop >= span.stop:
                span = next(spans)
        yield format_bio(part, type)


def io_spans(tokens, tags):
    previous = None
    start = None
    stop = None
    for token, tag in zip(tokens, tags):
        part, type = parse_bio(tag)
        # wikiner splits on I-PER B-PER for example
        if previous != type or part == B:
            if not previous and type:
                # O I
                start = token.start
            elif previous and type:
                # I-A I-B
                yield Span(start, stop, previous)
                start = token.start
            elif previous and not type:
                # I O
                yield Span(start, stop, previous)
                previous = None
        previous = type
        stop = token.stop
    if previous:
        yield Span(start, stop, previous)


#######
#
#   BIO
#
#########


def spans_bio(tokens, spans):
    spans = append_ellipsis(spans)
    span = next(spans)
    for token in tokens:
        part = O
        type = None
        if span:
            if token.start >= span.start:
                type = span.type
                if token.start == span.start:
                    part = B
                else:
                    part = I
            if token.stop >= span.stop:
                span = next(spans)
        yield format_bio(part, type)


def bio_spans(tokens, tags):
    previous = None
    start = None
    stop = None
    for token, tag in zip(tokens, tags):
        part, type = parse_bio(tag)
        if part == O:
            if previous:
                yield Span(start, stop, previous)
                previous = None
        elif part == B:
            if previous:
                yield Span(start, stop, previous)
            previous = type
            start = token.start
            stop = token.stop
        elif part == I:
            stop = token.stop
    if previous:
        yield Span(start, stop, previous)


#########
#
#   CONVERT
#
#########


def bio_io(tags):
    for tag in tags:
        part, type = parse_bio(tag)
        if part == B:
            part = I
        yield format_bio(part, type)


########
#
#   SELECT
#
######


def select_type_tags(tags, selected):
    for tag in tags:
        part, type = parse_bio(tag)
        if type != selected:
            part = O
            type = None
        yield format_bio(part, type)




import json
from gzip import (
    compress,
    decompress
)

import numpy as np

from slovnet.record import Record
from slovnet.tar import Tar, DumpTar
from slovnet.vocab import Vocab


PROTOCOL = 1

META = 'meta.json'
MODEL = 'model.json'


class Meta(Record):
    __attributes__ = ['id', 'protocol']

    def __init__(self, id, protocol=PROTOCOL):
        self.id = id
        self.protocol = protocol

    def check_protocol(self):
        if self.protocol != PROTOCOL:
            raise ValueError('Expected protocol=%r, got %r' % (PROTOCOL, self.protocol))


#######
#
#  ARRAY
#
#######


def array_name(id):
    return 'arrays/%d.bin' % id


def array_bytes(array):
    return array.tobytes()


def bytes_array(bytes, shape, dtype):
    return np.frombuffer(bytes, dtype).reshape(shape)


######
#
#  VOCAB
#
#######


def vocab_name(id):
    return 'vocabs/%s.gz' % id


def vocab_bytes(vocab):
    content = '\n'.join(vocab.items)
    bytes = content.encode('utf8')
    return compress(bytes)


def bytes_vocab(bytes):
    content = decompress(bytes).decode('utf8')
    items = content.splitlines()
    return Vocab(items)


######
#
#  PACK
#
########


def json_bytes(data):
    content = json.dumps(data, ensure_ascii=False, indent=2)
    return content.encode('utf8')


def bytes_json(bytes):
    return json.loads(bytes.decode('utf8'))


class Pack(Tar):
    def load_record(self, name, Record):
        bytes = self.read(name)
        data = bytes_json(bytes)
        return Record.from_json(data)

    def load_meta(self):
        return self.load_record(META, Meta)

    def load_model(self, Model):
        return self.load_record(MODEL, Model)

    def load_arrays(self, weights):
        for weight in weights:
            if not weight.is_id:
                continue

            shape, dtype, id = weight
            name = array_name(id)
            bytes = self.read(name)
            yield id, bytes_array(bytes, shape, dtype)

    def load_vocab(self, id):
        name = vocab_name(id)
        bytes = self.read(name)
        return bytes_vocab(bytes)


class DumpPack(DumpTar):
    def dump_record(self, record, name):
        bytes = json_bytes(record.as_json)
        self.write(bytes, name)

    def dump_meta(self, meta):
        self.dump_record(meta, META)

    def dump_model(self, model):
        self.dump_record(model, MODEL)

    def dump_arrays(self, arrays):
        for id, array in arrays.items():
            name = array_name(id)
            bytes = array_bytes(array)
            self.write(bytes, name)

    def dump_vocab(self, vocab, id):
        name = vocab_name(id)
        bytes = vocab_bytes(vocab)
        self.write(bytes, name)




import re


def bert_chunks(text):
    # diff with bert tokenizer 28 / 10000 ~0.3%
    # школа №3 -> школа, №3
    # @diet_prada -> @, diet, _, prada
    return re.findall(r'\w+|[^\w\s]', text)


def wordpiece(text, vocab, prefix='##'):
    start = 0
    stop = size = len(text)
    subs = []
    while start < size:
        sub = text[start:stop]
        if start > 0:
            sub = prefix + sub
        if sub in vocab.item_ids:
            subs.append(sub)
            start = stop
            stop = size
        else:
            stop -= 1
            if stop < start:
                return
    return subs


def safe_wordpiece(text, vocab):
    subs = wordpiece(text, vocab)
    if not subs:
        return [text]
    return subs


def bert_subs(text, vocab):
    return [
        sub
        for chunk in bert_chunks(text)
        for sub in safe_wordpiece(chunk, vocab)
    ]
